# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np

import sparse_ir
from sparse_ir import sampling


def test_decomp():
    rng = np.random.RandomState(4711)
    A = rng.randn(49, 39)

    Ad = sampling.DecomposedMatrix(A)
    norm_A = Ad.s[0] / Ad.s[-1]
    np.testing.assert_allclose(A, np.asarray(Ad), atol=1e-15 * norm_A, rtol=0)

    x = rng.randn(39)
    np.testing.assert_allclose(A @ x, Ad @ x, atol=1e-14 * norm_A, rtol=0)

    x = rng.randn(39, 3)
    np.testing.assert_allclose(A @ x, Ad @ x, atol=1e-14 * norm_A, rtol=0)

    y = rng.randn(49)
    np.testing.assert_allclose(np.linalg.lstsq(A, y, rcond=None)[0],
                               Ad.lstsq(y), atol=1e-14 * norm_A, rtol=0)

    y = rng.randn(49, 2)
    np.testing.assert_allclose(np.linalg.lstsq(A, y, rcond=None)[0],
                               Ad.lstsq(y), atol=1e-14 * norm_A, rtol=0)


def test_axis():
    rng = np.random.RandomState(4712)
    A = rng.randn(17, 21)

    Ad = sampling.DecomposedMatrix(A)
    norm_A = Ad.s[0] / Ad.s[-1]

    x = rng.randn(2, 21, 4, 7)
    ref = np.tensordot(A, x, (-1,1)).transpose((1,0,2,3))
    np.testing.assert_allclose(
            Ad.matmul(x, axis=1), ref,
            atol=1e-13 * norm_A, rtol=0)

def test_axis0():
    rng = np.random.RandomState(4712)
    A = rng.randn(17, 21)

    Ad = sampling.DecomposedMatrix(A)
    norm_A = Ad.s[0] / Ad.s[-1]

    x = rng.randn(21, 2)

    np.testing.assert_allclose(
            Ad.matmul(x, axis=0), A@x,
            atol=1e-13 * norm_A, rtol=0)

    np.testing.assert_allclose(
            Ad.matmul(x), A@x,
            atol=1e-13 * norm_A, rtol=0)


def test_tau_noise():
    basis = sparse_ir.IRBasis('F', 100)
    smpl = sparse_ir.TauSampling(basis)
    rng = np.random.RandomState(4711)

    rhol = basis.v([-.999, -.01, .5]) @ [0.8, -.2, 0.5]
    Gl = basis.s * rhol
    Gl_magn = np.linalg.norm(Gl)
    Gtau = smpl.evaluate(Gl)

    noise = 1e-5
    Gtau_n = Gtau +  noise * np.linalg.norm(Gtau) * rng.randn(*Gtau.shape)
    Gl_n = smpl.fit(Gtau_n)

    np.testing.assert_allclose(Gl, Gl_n, atol=12 * noise * Gl_magn, rtol=0)


def test_wn_noise():
    basis = sparse_ir.IRBasis('B', 99)
    smpl = sparse_ir.MatsubaraSampling(basis)
    rng = np.random.RandomState(4711)

    rhol = basis.v([-.999, -.01, .5]) @ [0.8, -.2, 0.5]
    Gl = basis.s * rhol
    Gl_magn = np.linalg.norm(Gl)
    Giw = smpl.evaluate(Gl)

    noise = 1e-5
    Giw_n = Giw +  noise * np.linalg.norm(Giw) * rng.randn(*Giw.shape)
    Gl_n = smpl.fit(Giw_n)
    np.testing.assert_allclose(Gl, Gl_n, atol=12 * noise * Gl_magn, rtol=0)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import warnings
import numpy as np

import scipy.linalg as sp_linalg
import numpy.polynomial.legendre as np_legendre


class Rule:
    """Quadrature rule.

    Approximation of an integral by a weighted sum over discrete points:

         ∫ f(x) * omega(x) * dx ~ sum(f(xi) * wi for (xi, wi) in zip(x, w))

    where we generally have superexponential convergence for smooth ``f(x)``
    with the number of quadrature points.
    """
    def __init__(self, x, w, x_forward=None, x_backward=None, a=-1, b=1):
        x = np.asarray(x)
        if x_forward is None:
            x_forward = x - a
        if x_backward is None:
            x_backward = b - x

        self.x = x
        self.w = np.asarray(w)
        self.x_forward = np.asarray(x_forward)
        self.x_backward = np.asarray(x_backward)
        self.a = a
        self.b = b

    def reseat(self, a, b):
        """Reseat current quadrature rule to new domain"""
        if self.a != -1 and self.b != 1:
            raise NotImplementedError("only supporting unscaled for now")

        extent = 0.5 * (b - a)
        midpoint = 0.5 * (b + a)
        x_new = extent * self.x
        x_new += midpoint
        x_forward = extent * self.x_forward
        x_backward = extent * self.x_backward
        w_new = extent * self.w
        return self.__class__(x_new, w_new, x_forward, x_backward, a, b)

    def scale(self, factor):
        """Scale weights by factor"""
        return self.__class__(self.x, self.w * factor, self.x_forward,
                              self.x_backward, self.a, self.b)

    def piecewise(self, edges):
        """Piecewise quadrature with the same quadrature rule, but scaled"""
        edges = np.asarray(edges)
        start = edges[:-1]
        stop = edges[1:]
        if not (stop > start).all():
            raise ValueError("segments ends must be ordered ascendingly")

        return self.join(*(self.reseat(start_i, stop_i)
                         for (start_i, stop_i) in zip(start, stop)))

    def astype(self, dtype):
        dtype = np.dtype(dtype)
        return self.__class__(
            self.x.astype(dtype), self.w.astype(dtype),
            self.x_forward.astype(dtype), self.x_backward.astype(dtype),
            dtype.type(self.a), dtype.type(self.b))

    @classmethod
    def join(cls, *gauss_list):
        """Join multiple Gauss quadratures together"""
        if not gauss_list:
            return cls((), ())

        a = gauss_list[0].a
        b = gauss_list[-1].b
        prev_b = a
        parts = []

        for curr in gauss_list:
            if curr.a != prev_b:
                raise ValueError("Gauss rules must be ascending")
            prev_b = curr.b
            x_forward = curr.x_forward + (curr.a - a)
            x_backward = curr.x_backward + (b - curr.b)
            parts.append((curr.x, curr.w, x_forward, x_backward))

        x, w, x_forward, x_backward = map(np.hstack, zip(*parts))
        return cls(x, w, x_forward, x_backward, a, b)


def legendre(n, dtype=float):
    """Gauss-Legendre quadrature"""
    return rule_from_recurrence(*_legendre_recurrence(n, dtype))


def legendre_collocation(rule, n=None):
    """Generate collocation matrix from Gauss-Legendre rule"""
    if n is None:
        n = rule.x.size

    res = np_legendre.legvander(rule.x, n - 1).T.copy()
    res *= rule.w

    invnorm = np.arange(0.5, n + 0.5, dtype=rule.x.dtype)
    res *= invnorm[:,None]
    return res


def rule_from_recurrence(alpha, beta, a, b):
    """Make new Gauss scheme based on recurrence coefficients.

    Given a set of polynomials ``P[n]`` defined by the following three-term
    recurrence relation::

        P[0](x)   == 1
        P[1](x)   == x - alpha[0]
        P[n+1](x) == (x - alpha[n]) * P[n] - beta[n] * P[n-1]

    we construct both a set of quadrature points ``x`` and weights ``w`` for
    Gaussian quadrature.  It is usually a good idea to work in extended
    precision for extra acccuracy in the quadrature rule.
    """
    dtype = np.result_type(alpha, beta)

    # First approximation of roots by finding eigenvalues of tridiagonal system
    # corresponding to the recursion
    beta[0] = b - a
    beta_is_pos = beta >= 0
    if not beta_is_pos.all():
        raise NotImplementedError("scipy solver cannot handle complex")

    sqrt_beta = np.sqrt(beta[1:])
    x = sp_linalg.eigvalsh_tridiagonal(alpha, sqrt_beta)
    x = x.astype(dtype)

    # These roots are usually only accurate to 100 ulps or so, so we improve
    # on them using a few iterations of the Newton method.
    prevdiff = 1.0
    maxiter = 5
    for _ in range(maxiter):
        p, dp, _, _ = _polyvalderiv(x, alpha, beta)
        diff = p / dp
        x -= diff

        # check convergence without relying on ATOL
        currdiff = np.abs(diff).max()
        #print(currdiff)
        if not (2 * currdiff <= prevdiff):
            break
        prevdiff = currdiff
    else:
        warnings.warn("Newton iteration did not converge, error = {.2g}"
                      .format(currdiff))

    # Now we know that the weights are proportional to the following:
    _, dp1, p0, _ = _polyvalderiv(x, alpha, beta)
    with np.errstate(over='ignore'):
        w = 1 / (dp1 * p0)
    w *= beta[0] / w.sum(initial=dtype.type(0))
    return Rule(x, w, x - a, b - x, a, b)


def _polyvalderiv(x, alpha, beta):
    """Return value and derivative of polynomial.

    Given a set of polynomials ``P[n]`` defined by a three-term recurrence,
    we evaluate both value and derviative for the highest polynomial and
    the second highest one.
    """
    n = len(alpha)
    p0 = np.ones_like(x)
    p1 = x - alpha[0] * p0
    dp0 = np.zeros_like(x)
    dp1 = p0
    for k in range(1, n):
        x_minus_alpha = x - alpha[k]
        p2 = x_minus_alpha * p1 - beta[k] * p0
        dp2 = p1 + x_minus_alpha * dp1 - beta[k] * dp0
        p0 = p1
        p1 = p2
        dp0 = dp1
        dp1 = dp2

    return p1, dp1, p0, dp0


def _legendre_recurrence(n, dtype=float):
    """Returns the alpha, beta for Gauss-Legendre integration"""
    # The Legendre polynomials are defined by the following recurrence:
    #
    #     (n + 1) * P[n+1](x) == (2 * n + 1) * x * P[n](x) - n * P[n-1](x)
    #
    # To normalize this, we realize that the prefactor of the highest power
    # of P[n] is (2n -1)!! / n!, which we divide by to obtain the "scaled"
    # beta values.
    dtype = np.dtype(dtype)
    k = np.arange(n, dtype=dtype)
    ksq = k**2
    alpha = np.zeros_like(k)
    beta = ksq / (4 * ksq - 1)
    beta[0] = 2
    one = dtype.type(1)
    return alpha, beta, -one, one



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
from warnings import warn


class KernelBase:
    """Integral kernel ``K(x, y)``.

    Abstract base class for an integral kernel, i.e., a real binary function
    ``K(x, y)`` used in a Fredhold integral equation of the first kind:

                        u(x) = ∫ K(x, y) v(y) dy

    where ``x ∈ [xmin, xmax]`` and ``y ∈ [ymin, ymax]``.  For its SVE to exist,
    the kernel must be square-integrable, for its singular values to decay
    exponentially, it must be smooth.
    """
    def __call__(self, x, y, x_plus=None, x_minus=None):
        """Evaluate kernel at point (x, y)

        For given ``x, y``, return the value of ``K(x, y)``. The arguments may
        be numpy arrays, in which case the function shall be evaluated over
        the broadcast arrays.

        The parameters ``x_plus`` and ``x_minus``, if given, shall contain the
        values of ``x - xmin`` and ``xmax - x``, respectively.  This is useful
        if either difference is to be formed and cancellation expected.
        """
        raise NotImplementedError()

    def sve_hints(self, eps):
        """Provide discretisation hints for the SVE routines.

        Advises the SVE routines of discretisation parameters suitable in
        tranforming the (infinite) SVE into an (finite) SVD problem.

        See: :class:``SVEHintsBase``.
        """
        raise NotImplementedError()

    @property
    def xrange(self):
        """Tuple ``(xmin, xmax)`` delimiting the range of allowed x values"""
        return -1, 1

    @property
    def yrange(self):
        """Tuple ``(ymin, ymax)`` delimiting the range of allowed y values"""
        return -1, 1

    @property
    def is_centrosymmetric(self):
        """Kernel is centrosymmetric.

        Returns true if and only if ``K(x, y) == K(-x, -y)`` for all values of
        ``x`` and ``y``.  This allows the kernel to be block-diagonalized,
        speeding up the singular value expansion by a factor of 4.  Defaults
        to false.
        """
        return False

    def get_symmetrized(self, sign):
        """Return symmetrized kernel ``K(x, y) + sign * K(x, -y)``.

        By default, this returns a simple wrapper over the current instance
        which naively performs the sum.  You may want to override if this
        to avoid cancellation.
        """
        return ReducedKernel(self, sign)

    @property
    def ypower(self):
        """Power with which the y coordinate scales."""
        return 0

    @property
    def conv_radius(self):
        """Convergence radius of the Matsubara basis asymptotic model.

        For improved relative numerical accuracy, the IR basis functions on the
        Matsubara axis ``basis.uhat(n)`` can be evaluated from an asymptotic
        expression for ``abs(n) > conv_radius``.  If ``conv_radius`` is
        ``None``, then the asymptotics are unused (the default).
        """
        return None


class SVEHintsBase:
    """Discretization hints for singular value expansion of a given kernel."""
    @property
    def segments_x(self):
        """Segments for piecewise polynomials on the ``x`` axis.

        List of segments on the ``x`` axis for the associated piecewise
        polynomial.  Should reflect the approximate position of roots of a
        high-order singular function in ``x``.
        """
        raise NotImplementedError()

    @property
    def segments_y(self):
        """Segments for piecewise polynomials on the ``x`` axis.

        List of segments on the ``y`` axis for the associated piecewise
        polynomial.  Should reflect the approximate position of roots of a
        high-order singular function in ``y``.
        """
        raise NotImplementedError()

    @property
    def ngauss(self):
        """Gauss-Legendre order to use to guarantee accuracy"""
        raise NotImplementedError()

    @property
    def nsvals(self):
        """Upper bound for number of singular values

        Upper bound on the number of singular values above the given
        threshold, i.e., where ``s[l] >= eps * s[0]```
        """
        raise NotImplementedError()


class KernelFFlat(KernelBase):
    """Fermionic analytical continuation kernel.

    In dimensionless variables ``x = 2*τ/β - 1``, ``y = β*ω/Λ``, the fermionic
    integral kernel is a function on ``[-1, 1] x [-1, 1]``::

        K(x, y) == exp(-Λ * y * (x + 1)/2) / (1 + exp(-Λ*y))
    """
    def __init__(self, lambda_):
        self.lambda_ = lambda_

    def __call__(self, x, y, x_plus=None, x_minus=None):
        x, y = _check_domain(self, x, y)
        u_plus, u_minus, v = _compute_uv(self.lambda_, x, y, x_plus, x_minus)
        return self._compute(u_plus, u_minus, v)

    def sve_hints(self, eps):
        return _SVEHintsFFlat(self, eps)

    def _compute(self, u_plus, u_minus, v):
        # By introducing u_\pm = (1 \pm x)/2 and v = lambda * y, we can write
        # the kernel in the following two ways:
        #
        #    k = exp(-u_+ * v) / (exp(-v) + 1)
        #      = exp(-u_- * -v) / (exp(v) + 1)
        #
        # We need to use the upper equation for v >= 0 and the lower one for
        # v < 0 to avoid overflowing both numerator and denominator
        abs_v = np.abs(v)
        enum = np.exp(-abs_v * np.where(v > 0, u_plus, u_minus))
        denom = 1 + np.exp(-abs_v)
        return enum / denom

    @property
    def is_centrosymmetric(self):
        return True

    def get_symmetrized(self, sign):
        if sign == -1:
            return _KernelFFlatOdd(self, sign)
        return super().get_symmetrized(sign)

    @property
    def conv_radius(self): return 40 * self.lambda_


class _SVEHintsFFlat(SVEHintsBase):
    def __init__(self, kernel, eps):
        self.kernel = kernel
        self.eps = eps

    @property
    def ngauss(self): return 10 if self.eps >= 1e-8 else 16

    @property
    def segments_x(self):
        nzeros = max(int(np.round(15 * np.log10(self.kernel.lambda_))), 1)
        diffs = 1./np.cosh(.143 * np.arange(nzeros))
        zeros_pos = diffs.cumsum()
        zeros_pos /= zeros_pos[-1]
        return np.concatenate((-zeros_pos[::-1], [0], zeros_pos))

    @property
    def segments_y(self):
        # Zeros around -1 and 1 are distributed asymptotically identical
        leading_diffs = np.array([
            0.01523, 0.03314, 0.04848, 0.05987, 0.06703, 0.07028, 0.07030,
            0.06791, 0.06391, 0.05896, 0.05358, 0.04814, 0.04288, 0.03795,
            0.03342, 0.02932, 0.02565, 0.02239, 0.01951, 0.01699])

        nzeros = max(int(np.round(20 * np.log10(self.kernel.lambda_))), 2)
        if nzeros < 20:
            leading_diffs = leading_diffs[:nzeros]
        diffs = .25 / np.exp(.141 * np.arange(nzeros))
        diffs[:leading_diffs.size] = leading_diffs
        zeros = diffs.cumsum()
        zeros = zeros[:-1] / zeros[-1]
        zeros -= 1
        return np.concatenate(([-1], zeros, [0], -zeros[::-1], [1]))

    @property
    def nsvals(self):
        log10_lambda = max(1, np.log10(self.kernel.lambda_))
        return int(np.round((25 + log10_lambda) * log10_lambda))


class KernelBFlat(KernelBase):
    """Bosonic analytical continuation kernel.

    In dimensionless variables ``x = 2*τ/β - 1``, ``y = β*ω/Λ``, the fermionic
    integral kernel is a function on ``[-1, 1] x [-1, 1]``::

        K(x, y) == y exp(-Λ * y * (x + 1)/2) / (exp(-Λ * y) - 1)

    Care has to be taken in evaluating this expression around ``y == 0``.
    """
    def __init__(self, lambda_):
        self.lambda_ = lambda_

    def __call__(self, x, y, x_plus=None, x_minus=None):
        x, y = _check_domain(self, x, y)
        u_plus, u_minus, v = _compute_uv(self.lambda_, x, y, x_plus, x_minus)
        return self._compute(u_plus, u_minus, v)

    def _compute(self, u_plus, u_minus, v):
        # With "reduced variables" u, v we have:
        #
        #   K = -1/lambda * exp(-u_+ * v) * v / (exp(-v) - 1)
        #     = -1/lambda * exp(-u_- * -v) * (-v) / (exp(v) - 1)
        #
        # where we again need to use the upper equation for v >= 0 and the
        # lower one for v < 0 to avoid overflow.
        abs_v = np.abs(v)
        enum = np.exp(-abs_v * np.where(v >= 0, u_plus, u_minus))
        dtype = v.dtype

        # The expression ``v / (exp(v) - 1)`` is tricky to evaluate: firstly,
        # it has a singularity at v=0, which can be cured by treating that case
        # separately.  Secondly, the denominator loses precision around 0 since
        # exp(v) = 1 + v + ..., which can be avoided using expm1(...)
        not_tiny = abs_v >= 1e-200
        denom = np.ones_like(abs_v)
        np.divide(abs_v, np.expm1(-abs_v, where=not_tiny),
                  out=denom, where=not_tiny)
        return -1/dtype.type(self.lambda_) * enum * denom

    def sve_hints(self, eps):
        return _SVEHintsBFlat(self, eps)

    @property
    def is_centrosymmetric(self):
        return True

    def get_symmetrized(self, sign):
        if sign == -1:
            return _KernelBFlatOdd(self, sign)
        return super().get_symmetrized(sign)

    @property
    def ypower(self): return 1

    @property
    def conv_radius(self): return 40 * self.lambda_


class _SVEHintsBFlat(SVEHintsBase):
    def __init__(self, kernel, eps):
        self.kernel = kernel
        self.eps = eps

    @property
    def ngauss(self): return 10 if self.eps >= 1e-8 else 16

    @property
    def segments_x(self):
        # Somewhat less accurate ...
        nzeros = 15 * np.log10(self.kernel.lambda_)
        diffs = 1./np.cosh(.18 * np.arange(nzeros))
        zeros_pos = diffs.cumsum()
        zeros_pos /= zeros_pos[-1]
        return np.concatenate((-zeros_pos[::-1], [0], zeros_pos))

    @property
    def segments_y(self):
        # Zeros around -1 and 1 are distributed asymptotically identical
        leading_diffs = np.array([
            0.01363, 0.02984, 0.04408, 0.05514, 0.06268, 0.06679, 0.06793,
            0.06669, 0.06373, 0.05963, 0.05488, 0.04987, 0.04487, 0.04005,
            0.03553, 0.03137, 0.02758, 0.02418, 0.02115, 0.01846])

        nzeros = max(int(np.round(20 * np.log10(self.kernel.lambda_))), 20)
        i = np.arange(nzeros)
        diffs = .12/np.exp(.0337 * i * np.log(i+1))
        #diffs[:leading_diffs.size] = leading_diffs
        zeros = diffs.cumsum()
        zeros = zeros[:-1] / zeros[-1]
        zeros -= 1
        return np.concatenate(([-1], zeros, [0], -zeros[::-1], [1]))

    @property
    def nsvals(self):
        log10_lambda = max(1, np.log10(self.kernel.lambda_))
        return int(28 * log10_lambda)


class ReducedKernel(KernelBase):
    """Restriction of centrosymmetric kernel to positive interval.

    For a kernel ``K`` on ``[-1, 1] x [-1, 1]`` that is centrosymmetrix, i.e.,
    ``K(x, y) == K(-x, -y)``, it is straight-forward to show that the left/right
    singular vectors can be chosen as either odd or even functions.

    Consequentially, they are singular functions of a reduced kernel ``K_red``
    on ``[0, 1] x [0, 1]`` that is given as either::

        K_red(x, y) == K(x, y) + sign * K(x, -y)

    This kernel is what this class represents.  The full singular functions can
    be reconstructed by (anti-)symmetrically continuing them to the negative
    axis.
    """
    def __init__(self, inner, sign=1):
        if not inner.is_centrosymmetric:
            raise ValueError("inner kernel must be centrosymmetric")
        if np.abs(sign) != 1:
            raise ValueError("sign must square to one")

        self.inner = inner
        self.sign = sign

    def __call__(self, x, y, x_plus=None, x_minus=None):
        x, y = _check_domain(self, x, y)

        # The reduced kernel is defined only over the interval [0, 1], which
        # means we must add one to get the x_plus for the inner kernels.  We
        # can compute this as 1 + x, since we are away from -1.
        x_plus = 1 + x_plus

        K_plus = self.inner(x, y, x_plus, x_minus)
        K_minus = self.inner(x, -y, x_plus, x_minus)
        return K_plus + K_minus if self.sign == 1 else K_plus - K_minus

    @property
    def xrange(self):
        _, xmax = self.inner.xrange
        return 0, xmax

    @property
    def yrange(self):
        _, ymax = self.inner.yrange
        return 0, ymax

    def sve_hints(self, eps):
        return _SVEHintsReduced(self.inner.sve_hints(eps))

    @property
    def is_centrosymmetric(self):
        """True iff K(x,y) = K(-x, -y)"""
        return False

    def get_symmetrized(self, sign):
        raise RuntimeError("cannot symmetrize twice")

    @property
    def ypower(self): return self.inner.ypower

    @property
    def conv_radius(self): return self.inner.conv_radius


class _SVEHintsReduced(SVEHintsBase):
    def __init__(self, inner_hints):
        self.inner_hints = inner_hints

    @property
    def ngauss(self): return self.inner_hints.ngauss

    @property
    def segments_x(self): return _symm_segments(self.inner_hints.segments_x)

    @property
    def segments_y(self): return _symm_segments(self.inner_hints.segments_y)

    @property
    def nsvals(self): return (self.inner_hints.nsvals + 1) // 2


class _KernelFFlatOdd(ReducedKernel):
    """Fermionic analytical continuation kernel, odd.

    In dimensionless variables ``x = 2*τ/β - 1``, ``y = β*ω/Λ``, the fermionic
    integral kernel is a function on ``[-1, 1] x [-1, 1]``::

        K(x, y) == -sinh(Λ/2 * x * y) / cosh(Λ/2 * y)
    """
    def __call__(self, x, y, x_plus=None, x_minus=None):
        result = super().__call__(x, y, x_plus, x_minus)

        # For x * y around 0, antisymmetrization introduces cancellation, which
        # reduces the relative precision.  To combat this, we replace the
        # values with the explicit form
        v_half = self.inner.lambda_/2 * y
        xy_small = x * v_half < 1
        cosh_finite = v_half < 85
        np.divide(-np.sinh(v_half * x, where=xy_small),
                  np.cosh(v_half, where=cosh_finite),
                  out=result, where=np.logical_and(xy_small, cosh_finite))
        return result


class _KernelBFlatOdd(ReducedKernel):
    """Bosonic analytical continuation kernel, odd.

    In dimensionless variables ``x = 2*τ/β - 1``, ``y = β*ω/Λ``, the fermionic
    integral kernel is a function on ``[-1, 1] x [-1, 1]``::

            K(x, y) = -y * sinh(Λ/2 * x * y) / sinh(Λ/2 * y)
    """
    def __call__(self, x, y, x_plus=None, x_minus=None):
        result = super().__call__(x, y, x_plus, x_minus)

        # For x * y around 0, antisymmetrization introduces cancellation, which
        # reduces the relative precision.  To combat this, we replace the
        # values with the explicit form
        v_half = self.inner.lambda_/2 * y
        xv_half = x * v_half
        xy_small = xv_half < 1
        sinh_range = np.logical_and(v_half > 1e-200, v_half < 85)
        np.divide(
            np.multiply(-y, np.sinh(xv_half, where=xy_small), where=xy_small),
            np.sinh(v_half, where=sinh_range),
            out=result, where=np.logical_and(xy_small, sinh_range))
        return result


def matrix_from_gauss(kernel, gauss_x, gauss_y):
    """Compute matrix for kernel from Gauss rule"""
    # (1 +- x) is problematic around x = -1 and x = 1, where the quadrature
    # nodes are clustered most tightly.  Thus we have the need for the
    # matrix method.
    return kernel(gauss_x.x[:,None], gauss_y.x[None,:],
                  gauss_x.x_forward[:,None], gauss_x.x_backward[:,None])


def _check_domain(kernel, x, y):
    """Check that arguments lie within the correct domain"""
    x = np.asarray(x)
    xmin, xmax = kernel.xrange
    if not (x >= xmin).all() or not (x <= xmax).all():
        raise ValueError("x values not in range [{:g},{:g}]".format(xmin, xmax))

    y = np.asarray(y)
    ymin, ymax = kernel.yrange
    if not (y >= ymin).all() or not (y <= ymax).all():
        raise ValueError("y values not in range [{:g},{:g}]".format(ymin, ymax))
    return x, y


def _symm_segments(x):
    x = np.asarray(x)
    if not np.allclose(x, -x[::-1]):
        raise ValueError("segments must be symmetric")
    xpos = x[x.size // 2:]
    if xpos[0] != 0:
        xpos = np.hstack([0, xpos])
    return xpos


def _compute_uv(lambda_, x, y, x_plus=None, x_minus=None):
    if x_plus is None:
        x_plus = 1 + x
    if x_minus is None:
        x_minus = 1 - x
    u_plus = .5 * x_plus
    u_minus = .5 * x_minus
    v = lambda_ * y
    return u_plus, u_minus, v



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import sparse_ir
from sparse_ir import sve
from sparse_ir import kernel
from sparse_ir import composite
from sparse_ir import augment

import pytest

@pytest.fixture(scope="module")
def basis():
    return sve.compute(kernel.KernelFFlat(42))

def _check_composite_poly(u_comp, u_list, test_points):
    assert u_comp.size == np.sum([u.size for u in u_list])
    assert u_comp.shape == (u_comp.size,)
    np.testing.assert_allclose(u_comp(test_points),
                               np.vstack([u(test_points) for u in u_list]))
    idx = 0
    for isub in range(len(u_list)):
        for ip in range(u_list[isub].size):
            np.testing.assert_allclose(u_comp[idx](test_points),
                                       u_list[isub][ip](test_points))
            idx += 1

def test_composite_poly(basis):
    u, s, v = basis
    l = s.size

    u_comp = composite.CompositeBasisFunction([u, u])
    _check_composite_poly(u_comp, [u, u], np.linspace(-1, 1, 10))

    uhat = u.hat("odd")
    uhat_comp = composite.CompositeBasisFunctionFT([uhat, uhat])
    _check_composite_poly(uhat_comp, [uhat, uhat], np.array([-3, 1, 5]))


def test_composite_basis():
    beta = 10
    wmax = 9.9
    basis = sparse_ir.FiniteTempBasis("F", beta, wmax, eps=1e-6)
    basis2 = sparse_ir.FiniteTempBasis("F", beta, wmax, eps=1e-3)
    basis_comp = composite.CompositeBasis([basis, basis2])
    _check_composite_poly(basis_comp.u, [basis.u, basis2.u],
                          np.linspace(0, beta, 10))
    _check_composite_poly(basis_comp.uhat, [basis.uhat, basis2.uhat],
                          np.array([1,3]))
    _check_composite_poly(basis_comp.v, [basis.v, basis2.v],
                          np.linspace(-wmax, -wmax, 10))


def test_augmented_bosonic_basis():
    """Augmented bosonic basis"""
    wmax = 2
    beta = 1000
    basis = sparse_ir.FiniteTempBasis("B", beta, wmax, eps=1e-6)
    basis_legg = augment.LegendreBasis("B", beta, 2)
    basis_comp = composite.CompositeBasis([basis_legg, basis])

    # G(tau) = c - e^{-tau*pole}/(1 - e^{-beta*pole})
    pole = 1.0
    c = 1e-2
    tau_smpl = sparse_ir.TauSampling(basis_comp)
    gtau = c - np.exp(-tau_smpl.sampling_points * pole)/(1 - np.exp(-beta * pole))
    gl_from_tau = tau_smpl.fit(gtau)

    gtau_reconst = tau_smpl.evaluate(gl_from_tau)
    np.testing.assert_allclose(gtau, gtau_reconst,
                               atol=1e-14 * np.abs(gtau).max(), rtol=0)


@pytest.mark.parametrize("stat", ["F", "B"])
def test_vertex_basis(stat):
    """Vertex basis"""
    wmax = 2
    beta = 1000
    basis = sparse_ir.FiniteTempBasis(stat, beta, wmax, eps=1e-6)
    basis_const = augment.MatsubaraConstBasis(stat, beta)
    basis_comp = composite.CompositeBasis([basis_const, basis])
    assert basis_comp.uhat is not None

    # G(iv) = c + 1/(iv-pole)
    pole = 1.0
    c = 1.0
    matsu_smpl = sparse_ir.MatsubaraSampling(basis_comp)
    giv = c  + 1/(1J*matsu_smpl.sampling_points * np.pi/beta - pole)
    gl = matsu_smpl.fit(giv)

    giv_reconst = matsu_smpl.evaluate(gl)

    np.testing.assert_allclose(giv, giv_reconst,
                               atol=np.abs(giv).max() * 1e-7, rtol=0)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
from _pytest.mark import param
import numpy as np
from sparse_ir import sve
from sparse_ir import kernel
from sparse_ir import poly

import pytest


@pytest.fixture(scope="module")
def basis():
    return sve.compute(kernel.KernelFFlat(42))


def test_shape(basis):
    u, s, v = basis
    l = s.size
    assert u.shape == (l,)

    assert u[3].shape == ()
    assert u[2:5].shape == (3,)


def test_eval(basis):
    u, s, v = basis
    l = s.size

    # evaluate
    np.testing.assert_array_almost_equal_nulp(
            u(0.4), [u[i](0.4) for i in range(l)])
    np.testing.assert_array_almost_equal_nulp(
            u([0.4, -0.2]),
            [[u[i](x) for x in (0.4, -0.2)] for i in range(l)])


def test_broadcast(basis):
    u, s, v = basis

    x = [0.3, 0.5]
    l = [2, 7]
    np.testing.assert_array_almost_equal_nulp(
            u.value(l, x), [u[ll](xx) for (ll, xx) in zip(l, x)])


def test_matrix_hat(basis):
    u, s, v = basis
    uhat = u.hat('odd')

    n = np.array([1, 3, 5, -1, -3, 5])
    result = uhat(n.reshape(3, 2))
    result_iter = uhat(n).reshape(-1, 3, 2)
    assert result.shape == result_iter.shape
    np.testing.assert_array_almost_equal_nulp(result, result_iter)


@pytest.mark.parametrize("lambda_, atol", [(42,1e-14), (1E+5,5e-12)])
def test_overlap(lambda_, atol):
    u, s, v = sve.compute(kernel.KernelFFlat(lambda_))

    # Keep only even number of polynomials
    u, s, v = u[:2*(s.size//2)], s[:2*(s.size//2)], v[:2*(s.size//2)]
    npoly = s.size

    np.testing.assert_allclose(u[0].overlap(u[0]), 1, rtol=0, atol=atol)

    ref = np.ones(s.size)
    ref[0] = 0
    np.testing.assert_allclose(
       np.abs(u.overlap(u[0])-1), ref, rtol=0, atol=atol
    )

    # Axis
    trans_f = lambda x: u[0](x).T
    np.testing.assert_allclose(
       np.abs(u.overlap(trans_f, axis=0)-1), ref, rtol=0, atol=atol
    )

    np.testing.assert_allclose(
        u.overlap(u, axis=-1), np.identity(s.size), rtol=0, atol=atol
    )

    u_tensor = poly.PiecewiseLegendrePoly(
                    u.data.reshape(u.data.shape[:2] + (npoly//2, 2)), u.knots)
    res = u_tensor.overlap(u_tensor, axis=-1)
    assert res.shape == (npoly//2, 2, npoly//2, 2)
    np.testing.assert_allclose(
        res.reshape(npoly, npoly), np.identity(s.size), rtol=0, atol=atol
    )


param_axis = []
shapes_axis = [(1,), (1,1), (1,2), (1,2,3)]
for shape in shapes_axis:
    for axis in range(len(shape)):
        param_axis.append((shape, axis))


@pytest.mark.parametrize("shape, axis", param_axis)
def test_overlap_axis(basis, shape, axis):
    u, s, v = basis
    def f(x):
        res = np.zeros((x.size, np.prod(shape)))
        for i in range(res.shape[1]):
            res[:,i] = i * x
        res = res.reshape((x.size,) + shape)
        return np.moveaxis(res, 0, axis)

    overlap = u.overlap(f, axis=axis)

    overlap_ref = np.empty((u.size, np.prod(shape)), dtype=overlap.dtype)
    for i in range(overlap_ref.shape[1]):
        overlap_ref[:,i] = u.overlap(lambda x: i*x)
    overlap_ref = overlap_ref.reshape((u.size,) + shape)

    np.testing.assert_allclose(overlap, overlap_ref, rtol=0,
                               atol=1e-10*np.abs(overlap_ref).max())



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import numpy.polynomial.legendre as np_legendre
import scipy.special as sp_special

from . import _roots
from . import gauss

try:
    from xprec import ddouble as _ddouble
except ImportError:
    _ddouble = None
    _xwork_dtype = float
else:
    _xwork_dtype = _ddouble


class PiecewiseLegendrePoly:
    """Piecewise Legendre polynomial.

    Models a function on the interval ``[-1, 1]`` as a set of segments on the
    intervals ``S[i] = [a[i], a[i+1]]``, where on each interval the function
    is expanded in scaled Legendre polynomials.
    """
    def __init__(self, data, knots, dx=None):
        """Piecewise Legendre polynomial"""
        if np.isnan(data).any():
            raise ValueError("PiecewiseLegendrePoly: data contains NaN!")
        if isinstance(knots, self.__class__):
            self.__dict__.update(knots.__dict__)
            self.data = data
            return

        data = np.array(data)
        knots = np.array(knots)
        polyorder, nsegments = data.shape[:2]
        if knots.shape != (nsegments+1,):
            raise ValueError("Invalid knots array")
        if not (knots[1:] >= knots[:-1]).all():
            raise ValueError("Knots must be monotonically increasing")
        if dx is None:
            dx = knots[1:] - knots[:-1]
        else:
            dx = np.array(dx)
            if not np.allclose(dx, knots[1:] - knots[:-1]):
                raise ValueError("dx must work with knots")

        self.nsegments = nsegments
        self.polyorder = polyorder
        self.xmin = knots[0]
        self.xmax = knots[-1]

        self.knots = knots
        self.dx = dx
        self.data = data
        self._xm = .5 * (knots[1:] + knots[:-1])
        self._inv_xs = 2/dx
        self._norm = np.sqrt(self._inv_xs)

    def __getitem__(self, l):
        """Return part of a set of piecewise polynomials"""
        if isinstance(l, tuple):
            new_data = self.data[(slice(None), slice(None), *l)]
        else:
            new_data = self.data[:,:,l]
        return self.__class__(new_data, self)

    def __call__(self, x):
        """Evaluate polynomial at position x"""
        i, xtilde = self._split(np.asarray(x))
        data = self.data[:, i]

        # Evaluate for all values of l.  x and data array must be
        # broadcast'able against each other, so we append dimensions here
        func_dims = self.data.ndim - 2
        datashape = i.shape + (1,) * func_dims
        res = np_legendre.legval(xtilde.reshape(datashape), data, tensor=False)
        res *= self._norm[i.reshape(datashape)]

        # Finally, exchange the x and vector dimensions
        order = tuple(range(i.ndim, i.ndim + func_dims)) + tuple(range(i.ndim))
        return res.transpose(*order)

    def value(self, l, x):
        """Return value for l and x."""
        if self.data.ndim != 3:
            raise ValueError("Only allowed for vector of data")

        l, x = np.broadcast_arrays(l, x)
        i, xtilde = self._split(x)
        data = self.data[:, i, l]

        # This should now neatly broadcast against each other
        res = np_legendre.legval(xtilde, data, tensor=False)
        res *= self._norm[i]
        return res

    def overlap(self, f, axis=None, _deg=None):
        r"""Evaluate overlap integral of this polynomial with function ``f``.

        Given the function ``f``, evaluate the integral::

            ∫ dx * f(x) * self(x)

        using piecewise Gauss-Legendre quadrature, where ``self`` are the
        polynomials.

        Arguments:
            f (callable):
                function that is called with a point ``x`` and returns ``f(x)``
                at that position.  If the ``axis`` argument is given, ``f``
                must be vectorized.
            axis (int or None):
                If `None` (the default), `f` is called repeatedly for all
                quadrature points.  If `axis` is not None, `f`, when called
                for a vector `x` returns `f(x[i])` at the `i`-th position along
                the given axis.

        Return:
            array-like object with shape (poly_dims, f_dims)
            poly_dims are the shape of the polynomial and f_dims are those
            of the function f(x).
        """
        if _deg is None:
            _deg = 2*self.polyorder

        # Get Gauss rule
        rule = gauss.legendre(_deg, self.data.dtype).piecewise(self.knots)
        x = rule.x

        # Multiply weights by polynomial at value
        pw = self(x) * rule.w
        fx = _apply_along_axis(f, x, axis)

        # Perform the summation and reshape the result
        int_flat = pw.reshape(self.size, x.size) @ fx.reshape(x.size, -1)
        return np.asarray(int_flat).reshape(self.shape + fx.shape[1:])

    def deriv(self, n=1):
        """Get polynomial for the n'th derivative"""
        ddata = np_legendre.legder(self.data, n)

        _scale_shape = (1, -1) + (1,) * (self.data.ndim - 2)
        scale = self._inv_xs ** n
        ddata *= scale.reshape(_scale_shape)
        return self.__class__(ddata, self.knots, self.dx)

    def hat(self, freq, n_asymp=None):
        """Get Fourier transformed object"""
        return PiecewiseLegendreFT(self, freq, n_asymp)

    def roots(self, alpha=2):
        """Find all roots of the piecewise polynomial

        Assume that between each two knots (pieces) there are at most ``alpha``
        roots.
        """
        if self.data.ndim > 2:
            raise ValueError("select single polynomial before calling roots()")

        grid = _refine_grid(self.knots, alpha)
        return _roots.find_all(self, grid)

    @property
    def shape(self): return self.data.shape[2:]

    @property
    def size(self): return self.data[:1,:1].size

    @property
    def ndim(self): return self.data.ndim - 2

    def _in_domain(self, x):
        return (x >= self.xmin).all() and (x <= self.xmax).all()

    def _split(self, x):
        """Split segment"""
        if not self._in_domain(x):
            raise ValueError("x must be in [%g, %g]" % (self.xmin, self.xmax))

        i = self.knots.searchsorted(x, 'right').clip(None, self.nsegments)
        i -= 1
        xtilde = x - self._xm[i]
        xtilde *= self._inv_xs[i]
        return i, xtilde


class PiecewiseLegendreFT:
    """Fourier transform of a piecewise Legendre polynomial.

    For a given frequency index ``n``, the Fourier transform of the Legendre
    function is defined as::

            phat(n) == ∫ dx exp(1j * pi * n * x / (xmax - xmin)) p(x)

    The polynomial is continued either periodically (``freq='even'``), in which
    case ``n`` must be even, or antiperiodically (``freq='odd'``), in which case
    ``n`` must be odd.
    """
    _DEFAULT_GRID = np.hstack([np.arange(2**6),
                        (2**np.linspace(6, 25, 16*(25-6)+1)).astype(int)])

    def __init__(self, poly, freq='even', n_asymp=None):
        if poly.xmin != -1 or poly.xmax != 1:
            raise NotImplementedError("Only interval [-1, 1] supported")
        self.poly = poly
        self.freq = freq
        self.zeta = {'any': None, 'even': 0, 'odd': 1}[freq]
        if n_asymp is None:
            self.n_asymp = np.inf
            self._model = None
        else:
            self.n_asymp = n_asymp
            self._model = _power_model(freq, poly)

    @property
    def shape(self): return self.poly.shape

    @property
    def size(self): return self.poly.size

    @property
    def ndim(self): return self.poly.ndim

    def __getitem__(self, l):
        return self.__class__(self.poly[l], self.freq)

    def __call__(self, n):
        """Obtain Fourier transform of polynomial for given frequencies"""
        n = self._check_domain(n)
        n_flat = n.ravel()
        result_flat = _compute_unl_inner(self.poly, n_flat)

        # We use use the asymptotics at frequencies larger than conv_radius
        # since it has lower relative error.
        cond_outer = np.abs(n_flat) >= self.n_asymp
        if cond_outer.any():
            n_outer = n_flat[cond_outer]
            result_flat[..., cond_outer] = self._model.giw(n_outer).T

        return result_flat.reshape(self.poly.shape + n.shape)

    def extrema(self, part=None, grid=None):
        """Obtain extrema of fourier-transformed polynomial."""
        if self.poly.shape:
            raise ValueError("select single polynomial")
        if grid is None:
            grid = self._DEFAULT_GRID

        f = self._func_for_part(part)
        x0 = _roots.discrete_extrema(f, PiecewiseLegendreFT._DEFAULT_GRID)
        x0 = 2 * x0 + self.zeta
        return self._symmetrize(x0)

    def _check_domain(self, n):
        n = np.asarray(n)
        if self.freq == 'any':
            return n
        if np.issubdtype(n.dtype, np.integer):
            nint = n
        else:
            nint = n.astype(int)
            if not (n == nint).all():
                raise ValueError("n must be integers")
        if not (nint % 2 == self.zeta).all():
            raise ValueError("n have wrong parity")
        return nint

    def _func_for_part(self, part=None):
        if part is None:
            parity = self.poly(self.poly.xmax) / self.poly(self.poly.xmin)
            if np.allclose(parity, 1):
                part = 'real' if self.zeta == 0 else 'imag'
            elif np.allclose(parity, -1):
                part = 'imag' if self.zeta == 0 else 'real'
            else:
                raise ValueError("cannot detect parity.")
        if part == 'real':
            return lambda n: self(2*n + self.zeta).real
        elif part == 'imag':
            return lambda n: self(2*n + self.zeta).imag
        else:
            raise ValueError("part must be either 'real' or 'imag'")

    def _symmetrize(self, x0):
        if x0[0] == 0:
            x0 = np.hstack([-x0[::-1], x0[1:]])
        else:
            x0 = np.hstack([-x0[::-1], x0])
        return x0


def _imag_power(n):
    """Imaginary unit raised to an integer power without numerical error"""
    n = np.asarray(n)
    if not np.issubdtype(n.dtype, np.integer):
        raise ValueError("expecting set of integers here")
    cycle = np.array([1, 0+1j, -1, 0-1j], complex)
    return cycle[n % 4]


def _get_tnl(l, w):
    r"""Fourier integral of the l-th Legendre polynomial::

        T_l(w) == \int_{-1}^1 dx \exp(iwx) P_l(x)
    """
    # spherical_jn gives NaN for w < 0, but since we know that P_l(x) is real,
    # we simply conjugate the result for w > 0 in these cases.
    result = 2 * _imag_power(l) * sp_special.spherical_jn(l, np.abs(w))
    np.conjugate(result, out=result, where=w < 0)
    return result


def _shift_xmid(knots, dx):
    r"""Return midpoint relative to the nearest integer plus a shift.

    Return the midpoints ``xmid`` of the segments, as pair ``(diff, shift)``,
    where shift is in ``(0,1,-1)`` and ``diff`` is a float such that
    ``xmid == shift + diff`` to floating point accuracy.
    """
    dx_half = dx / 2
    xmid_m1 = dx.cumsum() - dx_half
    xmid_p1 = -dx[::-1].cumsum()[::-1] + dx_half
    xmid_0 = knots[1:] - dx_half

    shift = np.round(xmid_0).astype(int)
    diff = np.choose(shift+1, (xmid_m1, xmid_0, xmid_p1))
    return diff, shift


def _phase_stable(poly, wn):
    """Phase factor for the piecewise Legendre to Matsubara transform.

    Compute the following phase factor in a stable way::

        np.exp(1j * np.pi/2 * wn[:,None] * poly.dx.cumsum()[None,:])
    """
    # A naive implementation is losing precision close to x=1 and/or x=-1:
    # there, the multiplication with `wn` results in `wn//4` almost extra turns
    # around the unit circle.  The cosine and sines will first map those
    # back to the interval [-pi, pi) before doing the computation, which loses
    # digits in dx.  To avoid this, we extract the nearest integer dx.cumsum()
    # and rewrite above expression like below.
    #
    # Now `wn` still results in extra revolutions, but the mapping back does
    # not cut digits that were not there in the first place.
    xmid_diff, extra_shift = _shift_xmid(poly.knots, poly.dx)

    if np.issubdtype(wn.dtype, np.integer):
        shift_arg = wn[None,:] * xmid_diff[:,None]
    else:
        delta_wn, wn = np.modf(wn)
        wn = wn.astype(int)
        shift_arg = wn[None,:] * xmid_diff[:,None]
        shift_arg += delta_wn[None,:] * (extra_shift + xmid_diff)[:,None]

    phase_shifted = np.exp(0.5j * np.pi * shift_arg)
    corr = _imag_power((extra_shift[:,None] + 1) * wn[None,:])
    return corr * phase_shifted


def _compute_unl_inner(poly, wn):
    """Compute piecewise Legendre to Matsubara transform."""
    dx_half = poly.dx / 2

    data_flat = poly.data.reshape(*poly.data.shape[:2], -1)
    data_sc = data_flat * np.sqrt(dx_half/2)[None,:,None]
    p = np.arange(poly.polyorder)

    wred = np.pi/2 * wn
    phase_wi = _phase_stable(poly, wn)
    t_pin = _get_tnl(p[:,None,None], wred[None,:] * dx_half[:,None]) * phase_wi

    # Perform the following, but faster:
    #   resulth = einsum('pin,pil->nl', t_pin, data_sc)
    npi = poly.polyorder * poly.nsegments
    result_flat = t_pin.reshape(npi,-1).T.dot(data_sc.reshape(npi,-1)).T
    return result_flat.reshape(*poly.data.shape[2:], wn.size)


class _PowerModel:
    """Model from a high-frequency series expansion::

        A(iw) == sum(A[n] / (iw)**(n+1) for n in range(1, N))

    where ``iw == 1j * pi/2 * wn`` is a reduced imaginary frequency, i.e.,
    ``wn`` is an odd/even number for fermionic/bosonic frequencies.
    """
    def __init__(self, moments):
        """Initialize model"""
        self.moments = np.asarray(moments)
        self.nmom, self.nl = self.moments.shape

    def _giw_ravel(self, wn):
        """Return model Green's function for vector of frequencies"""
        result_dtype = np.result_type(1j, wn, self.moments)
        result = np.zeros((wn.size, self.nl), result_dtype)
        inv_iw = 1j * np.pi/2 * wn
        np.reciprocal(inv_iw, out=inv_iw, where=(wn != 0))
        for mom in self.moments[::-1]:
            result += mom
            result *= inv_iw[:, None]
        return result

    def giw(self, wn):
        """Return model Green's function for reduced frequencies"""
        wn = np.array(wn)
        return self._giw_ravel(wn.ravel()).reshape(wn.shape + (self.nl,))


def _derivs(ppoly, x):
    """Evaluate polynomial and its derivatives at specific x"""
    yield ppoly(x)
    for _ in range(ppoly.polyorder-1):
        ppoly = ppoly.deriv()
        yield ppoly(x)


def _power_moments(stat, deriv_x1):
    """Return moments"""
    statsign = {'odd': -1, 'even': 1}[stat]
    mmax, lmax = deriv_x1.shape
    m = np.arange(mmax)[:,None]
    l = np.arange(lmax)[None,:]
    coeff_lm = ((-1.0)**(m+1) + statsign * (-1.0)**l) * deriv_x1
    return -statsign/np.sqrt(2.0) * coeff_lm


def _power_model(stat, poly):
    deriv_x1 = np.asarray(list(_derivs(poly, x=1)))
    if deriv_x1.ndim == 1:
        deriv_x1 = deriv_x1[:,None]
    moments = _power_moments(stat, deriv_x1)
    return _PowerModel(moments)


def _refine_grid(knots, alpha):
    """Linear refinement of grid"""
    result = np.linspace(knots[:-1], knots[1:], alpha, endpoint=False)
    return np.hstack((result.T.ravel(), knots[-1]))


def _apply_along_axis(f, x, axis=None):
    if axis is None:
        fx = [f(xi) for xi in x]
        fx = np.array(fx)
        if fx.dtype is np.dtype(object):
            raise ValueError("incompatible shapes")
    else:
        fx = np.asarray(f(x))
        if fx.shape[axis] != x.size:
            raise ValueError("inconsistent result shape")
        if axis != 0:
            fx = np.moveaxis(fx, axis, 0)
    return fx



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
from warnings import warn


class SamplingBase:
    """Base class for sparse sampling.

    Encodes the "basis transformation" of a propagator from the truncated IR
    basis coefficients ``G_ir[l]`` to time/frequency sampled on sparse points
    ``G(x[i])`` together with its inverse, a least squares fit::

             ________________                   ___________________
            |                |    evaluate     |                   |
            |     Basis      |---------------->|     Value on      |
            |  coefficients  |<----------------|  sampling points  |
            |________________|      fit        |___________________|


    Attributes:
        basis: IR Basis instance
        matrix: Evaluation matrix is decomposed form
        sampling_points: Set of sampling points
    """
    def __init__(self, basis, sampling_points):
        self.sampling_points = np.array(sampling_points)
        self.basis = basis
        self.matrix = DecomposedMatrix(
                        self.__class__.eval_matrix(basis, self.sampling_points))

        # Check conditioning
        self.cond = self.matrix.s[0] / self.matrix.s[-1]
        if self.cond > 1e8:
            warn("Sampling matrix is poorly conditioned (cond = %.2g)"
                 % self.cond, ConditioningWarning)

    def evaluate(self, al, axis=None):
        """Evaluate the basis coefficients at the sparse sampling points"""
        return self.matrix.matmul(al, axis)

    def fit(self, ax, axis=None):
        """Fit basis coefficients from the sparse sampling points"""
        return self.matrix.lstsq(ax, axis)

    @classmethod
    def eval_matrix(cls, basis, x):
        """Return evaluation matrix from coefficients to sampling points"""
        raise NotImplementedError()


class TauSampling(SamplingBase):
    """Sparse sampling in imaginary time.

    Allows the transformation between the IR basis and a set of sampling points
    in (scaled/unscaled) imaginary time.
    """
    def __init__(self, basis, sampling_points=None):
        if sampling_points is None:
            sampling_points = basis.default_tau_sampling_points()
        super().__init__(basis, sampling_points)

    @classmethod
    def eval_matrix(cls, basis, x):
        return basis.u(x).T

    @property
    def tau(self):
        """Sampling points in (reduced) imaginary time"""
        return self.sampling_points


class MatsubaraSampling(SamplingBase):
    """Sparse sampling in Matsubara frequencies.

    Allows the transformation between the IR basis and a set of sampling points
    in (scaled/unscaled) imaginary frequencies.
    """
    def __init__(self, basis, sampling_points=None):
        if sampling_points is None:
            sampling_points = basis.default_matsubara_sampling_points()
        super().__init__(basis, sampling_points)

    @classmethod
    def eval_matrix(cls, basis, x):
        return basis.uhat(x).T

    @property
    def wn(self):
        """Sampling points as (reduced) Matsubara frequencies"""
        return self.sampling_points


class DecomposedMatrix:
    """Matrix in SVD decomposed form for fast and accurate fitting.

    Stores a matrix ``A`` together with its thin SVD form::

        A == (u * s) @ vH.

    This allows for fast and accurate least squares fits using ``A.lstsq(x)``.
    """
    @classmethod
    def get_svd_result(cls, a, eps=None):
        """Construct decomposition from matrix"""
        u, s, vH = np.linalg.svd(a, full_matrices=False)
        where = s.astype(bool) if eps is None else s/s[0] <= eps
        if not where.all():
            return u[:, where], s[where], vH[where]
        else:
            return u, s, vH

    def __init__(self, a, svd_result=None):
        a = np.asarray(a)
        if a.ndim != 2:
            raise ValueError("a must be of matrix form")
        if svd_result is None:
            u, s, vH = self.__class__.get_svd_result(a)
        else:
            u, s, vH = map(np.asarray, svd_result)

        self._a = a
        self._uH = np.array(u.conj().T)
        self._s = s
        self._v = np.array(vH.conj().T)

    def __matmul__(self, x):
        """Matrix-matrix multiplication."""
        return self._a @ x

    def matmul(self, x, axis=None):
        """Compute ``A @ x`` (optionally along specified axis of x)"""
        return _matop_along_axis(self._a.__matmul__, x, axis)

    def _lstsq(self, x):
        r = self._uH @ x
        r = r / (self._s[:, None] if r.ndim > 1 else self._s)
        return self._v @ r

    def lstsq(self, x, axis=None):
        """Return ``y`` such that ``np.linalg.norm(A @ y - x)`` is minimal"""
        return _matop_along_axis(self._lstsq, x, axis)

    def __array__(self, dtype=""):
        """Convert to numpy array."""
        return self._a if dtype == "" else self._a.astype(dtype)

    @property
    def a(self):
        """Full matrix"""
        return self._a

    @property
    def u(self):
        """Left singular vectors, aranged column-wise"""
        return self._uH.conj().T

    @property
    def s(self):
        """Most significant, nonzero singular values"""
        return self._s

    @property
    def vH(self):
        """Right singular vectors, transposed"""
        return self._v.conj().T


class ConditioningWarning(RuntimeWarning):
    pass


def _matop_along_axis(op, x, axis=None):
    if axis is None:
        return op(x)

    x = np.asarray(x)
    target_axis = max(x.ndim - 2, 0)
    x = np.moveaxis(x, axis, target_axis)
    r = op(x)
    return np.moveaxis(r, target_axis, axis)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
"""
Drop-in replacement for the irbasis module.

This is designed to be a drop-in replacement for ``irbasis``, where the basis
can be computed on-the-fly for arbitrary values of Lambda.  In other words,
you should be able to replace ``irbasis`` with ``sparse_ir.adapter`` and
everything should hopefully still work.

Note however that on-the-fly computation typically has lower accuracy.  Thus,
by default we only populate the basis down to singular values of 1.5e-8.
You can change this by setting the `ACCURACY` parameter.
"""
from os import stat
import numpy as _np
from warnings import warn

from . import basis as _basis
from . import poly as _poly

try:
    import xprec as _xprec
except ImportError:
    ACCURACY = 1.0e-9
    WARN_ACCURACY = True
else:
    ACCURACY = 1.0e-15
    WARN_ACCURACY = False


def load(statistics, Lambda, h5file=None):
    if WARN_ACCURACY:
        warn("xprec package is not found - expect degraded accuracy!\n"
             "To squelch this warning, set WARN_ACCURACY to False.")

    basis = _basis.IRBasis(statistics, float(Lambda))
    return Basis(statistics, Lambda, basis.u, basis.s, basis.v)


class Basis:
    def __init__(self, statistics, Lambda, u, s, v):
        self._statistics = statistics
        self._Lambda = Lambda
        self._u = u
        self._s = s
        self._v = v

    @property
    def Lambda(self):
        """Dimensionless parameter of IR basis"""
        return self._Lambda

    @property
    def statistics(self):
        """Statistics, either "F" for fermions or "B" for bosons"""
        return self._statistics

    def dim(self):
        """Return dimension of basis"""
        return self._s.size

    def sl(self, l=None):
        """Return the singular value for the l-th basis function"""
        return _select(self._s, l)

    def ulx(self, l, x):
        """Return value of basis function for x"""
        return _selectvals(self._u, l, x)

    def d_ulx(self, l, x, order, section=None):
        """Return (higher-order) derivatives of u_l(x)"""
        return _selectvals(self._u.deriv(order), l, x)

    def vly(self, l, y):
        """Return value of basis function for y"""
        return _selectvals(self._v, l, y)

    def d_vly(self, l, y, order):
        """Return (higher-order) derivatives of v_l(y)"""
        return _selectvals(self._v.deriv(order), l, y)

    def compute_unl(self, n, whichl=None):
        """Compute transformation matrix from IR to Matsubara frequencies"""
        n = _np.ravel(n)
        freq = {'F': 'odd', 'B': 'even'}[self._statistics]
        uhat = _select(self._u, whichl).hat(freq)
        return _np.squeeze(uhat(2 * n + uhat.zeta).T)

    def num_sections_x(self):
        "Number of sections of piecewise polynomial representation of u_l(x)"
        return self._u.nsegments

    @property
    def section_edges_x(self):
        """End points of sections for u_l(x)"""
        return self._u.knots

    def num_sections_y(self):
        "Number of sections of piecewise polynomial representation of v_l(y)"
        return self._v.nsegments

    @property
    def section_edges_y(self):
        """End points of sections for v_l(y)"""
        return self._v.knots

    def sampling_points_x(self, whichl):
        """Computes "optimal" sampling points in x space for given basis"""
        return _poly.sampling_points_x(self._u[whichl])

    def sampling_points_y(self, whichl):
        """Computes "optimal" sampling points in y space for given basis"""
        return _poly.sampling_points_x(self._v[whichl])

    def sampling_points_matsubara(self, whichl):
        """Computes sampling points in Matsubara domain for given basis"""
        zeta = 0 if self.statistics == 'B' else 1
        part = ['real', 'imag'][(whichl + zeta) % 2]
        return _poly.find_hat_extrema(self._u[whichl], part, zeta)


def sampling_points_x(b, whichl):
    """Computes "optimal" sampling points in x space for given basis"""
    return b.sampling_points_x(whichl)


def sampling_points_y(b, whichl):
    """Computes "optimal" sampling points in y space for given basis"""
    return b.sampling_points_y(whichl)


def sampling_points_matsubara(b, whichl):
    """Computes "optimal" sampling points in Matsubara domain for given basis"""
    return b.sampling_points_matsubara(whichl)


def _select(p, l):
    return p if l is None else p[l]


def _selectvals(p, l, x):
    return p(x) if l is None else p.value(l, x)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import pytest

from sparse_ir import kernel
from sparse_ir import gauss

KERNELS = [
    kernel.KernelFFlat(9),
    kernel.KernelBFlat(8),
    kernel.KernelFFlat(120_000),
    kernel.KernelBFlat(127_500),
    kernel.KernelFFlat(40_000).get_symmetrized(-1),
    kernel.KernelBFlat(35_000).get_symmetrized(-1),
    ]


@pytest.mark.parametrize("K", KERNELS)
def test_accuracy(K):
    dtype = np.float32
    dtype_x = np.float64

    rule = gauss.legendre(10, dtype)
    hints = K.sve_hints(2.2e-16)
    gauss_x = rule.piecewise(hints.segments_x)
    gauss_y = rule.piecewise(hints.segments_y)
    eps = np.finfo(dtype).eps
    tiny = np.finfo(dtype).tiny / eps

    result = kernel.matrix_from_gauss(K, gauss_x, gauss_y)
    result_x = kernel.matrix_from_gauss(
                    K, gauss_x.astype(dtype_x), gauss_y.astype(dtype_x))
    magn = np.abs(result_x).max()
    np.testing.assert_allclose(result, result_x, atol = 2 * magn * eps, rtol=0,
                               err_msg="absolute precision tool poor")

    with np.errstate(invalid='ignore'):
        reldiff = np.where(np.abs(result_x) < tiny, 1, result / result_x)
    np.testing.assert_allclose(reldiff, 1, atol=100 * eps, rtol=0,
                               err_msg="relative precision too poor")



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
from warnings import warn

from . import kernel as _kernel
from . import sve


class IRBasis:
    """Intermediate representation (IR) basis in reduced variables.

    For a continuation kernel from real frequencies, ω ∈ [-ωmax, ωmax], to
    imaginary time, τ ∈ [0, β], this class stores the truncated singular
    value expansion or IR basis::

        K(x, y) ≈ sum(u[l](x) * s[l] * v[l](y) for l in range(L))

    The functions are given in reduced variables, ``x = 2*τ/β - 1`` and
    ``y = ω/ωmax``, which scales both sides to the interval ``[-1, 1]``.  The
    kernel then only depends on a cutoff parameter ``Λ = β * ωmax``.

    Example:
        The following example code assumes the spectral function is a single
        pole at x = 0.2::

            # Compute IR basis suitable for fermions and β*W <= 42
            import sparse_ir
            basis = sparse_ir.IRBasis(statistics='F', lambda_=42)

            # Assume spectrum is a single pole at x = 0.2, compute G(iw)
            # on the first few Matsubara frequencies
            gl = basis.s * basis.v(0.2)
            giw = gl @ basis.uhat([1, 3, 5, 7])

    See also:
        :class:`FiniteTempBasis` for a basis directly in time/frequency.

    Attributes:
        u (sparse_ir.poly.PiecewiseLegendrePoly):
            Set of IR basis functions on the reduced imaginary time (`x`) axis.
            These functions are stored as piecewise Legendre polynomials.

            To obtain the value of all basis functions at a point or a array of
            points `x`, you can call the function ``u(x)``.  To obtain a single
            basis function, a slice or a subset `l`, you can use ``u[l]``.

        uhat (sparse_ir.poly.PiecewiseLegendreFT):
            Set of IR basis functions on the Matsubara frequency (`wn`) axis.
            These objects are stored as a set of Bessel functions.

            To obtain the value of all basis functions at a Matsubara frequency
            or a array of points `wn`, you can call the function ``uhat(wn)``.
            Note that we expect reduced frequencies, which are simply even/odd
            numbers for bosonic/fermionic objects. To obtain a single basis
            function, a slice or a subset `l`, you can use ``uhat[l]``.

        s:
            vector of singular values of the continuation kernel

        v (sparse_ir.poly.PiecewiseLegendrePoly):
            Set of IR basis functions on the reduced real frequency (`y`) axis.
            These functions are stored as piecewise Legendre polynomials.

            To obtain the value of all basis functions at a point or a array of
            points `y`, you can call the function ``v(y)``.  To obtain a single
            basis function, a slice or a subset `l`, you can use ``v[l]``.
    """
    def __init__(self, statistics, lambda_, eps=None, *, kernel=None,
                 sve_result=None):
        if not (lambda_ >= 0):
            raise ValueError("kernel cutoff lambda must be non-negative")

        self.kernel = _get_kernel(statistics, lambda_, kernel)
        if sve_result is None:
            u, s, v = sve.compute(self.kernel, eps)
        else:
            u, s, v = sve_result
            if u.shape != s.shape or s.shape != v.shape:
                raise ValueError("mismatched shapes in SVE")

        if eps is None and sve_result is None and not sve.HAVE_XPREC:
            warn("xprec package is not available:\n"
                 "expect single precision (1.5e-8) only as both cutoff and\n"
                 "accuracy of the basis functions")

        # The radius of convergence of the asymptotic expansion is Lambda/2,
        # so for significantly larger frequencies we use the asymptotics,
        # since it has lower relative error.
        _even_odd = {'F': 'odd', 'B': 'even'}[statistics]
        self.u = u
        self.uhat = u.hat(_even_odd, n_asymp=self.kernel.conv_radius)
        self.s = s
        self.v = v
        self.statistics = statistics

    def __getitem__(self, index):
        """Return basis functions/singular values for given index/indices.

        This can be used to truncate the basis to the n most significant
        singular values: `basis[:3]`.
        """
        sve_result = self.u[index], self.s[index], self.v[index]
        return self.__class__(self.kernel, self.statistics, sve_result=sve_result)

    @property
    def lambda_(self):
        """Basis cutoff parameter Λ = β * ωmax"""
        return self.kernel.lambda_

    @property
    def size(self):
        """Number of basis functions / singular values."""
        return self.u.size

    @property
    def shape(self):
        """Shape of the basis function set"""
        return self.u.shape

    @property
    def beta(self):
        """Inverse temperature (this is `None` because unscaled basis)"""
        return None

    @property
    def wmax(self):
        """Frequency cutoff (this is `None` because unscaled basis)"""
        return None

    @property
    def sve_result(self):
        return self.u, self.s, self.v

    def default_tau_sampling_points(self):
        """Default sampling points on the imaginary time/x axis"""
        return _default_tau_sampling_points(self.u)

    def default_matsubara_sampling_points(self, *, mitigate=True):
        """Default sampling points on the imaginary frequency axis"""
        return _default_matsubara_sampling_points(self.uhat, mitigate)


class FiniteTempBasis:
    """Intermediate representation (IR) basis for given temperature.

    For a continuation kernel from real frequencies, ω ∈ [-wmax, wmax], to
    imaginary time, τ ∈ [0, beta], this class stores the truncated singular
    value expansion or IR basis::

        K(τ, ω) ≈ sum(u[l](τ) * s[l] * v[l](ω) for l in range(L))

    This basis is inferred from a reduced form by appropriate scaling of
    the variables.

    Example:
        The following example code assumes the spectral function is a single
        pole at ω = 2.5::

            # Compute IR basis for fermions and β = 10, W <= 4.2
            import sparse_ir
            basis = sparse_ir.FiniteTempBasis(statistics='F', beta=10, wmax=4.2)

            # Assume spectrum is a single pole at ω = 2.5, compute G(iw)
            # on the first few Matsubara frequencies
            gl = basis.s * basis.v(2.5)
            giw = gl @ basis.uhat([1, 3, 5, 7])

    Attributes:
        u (sparse_ir.poly.PiecewiseLegendrePoly):
            Set of IR basis functions on the imaginary time (`tau`) axis.
            These functions are stored as piecewise Legendre polynomials.

            To obtain the value of all basis functions at a point or a array of
            points `x`, you can call the function ``u(x)``.  To obtain a single
            basis function, a slice or a subset `l`, you can use ``u[l]``.

        uhat (sparse_ir.poly.PiecewiseLegendreFT):
            Set of IR basis functions on the Matsubara frequency (`wn`) axis.
            These objects are stored as a set of Bessel functions.

            To obtain the value of all basis functions at a Matsubara frequency
            or a array of points `wn`, you can call the function ``uhat(wn)``.
            Note that we expect reduced frequencies, which are simply even/odd
            numbers for bosonic/fermionic objects. To obtain a single basis
            function, a slice or a subset `l`, you can use ``uhat[l]``.

        s:
            vector of singular values of the continuation kernel

        v (sparse_ir.poly.PiecewiseLegendrePoly):
            Set of IR basis functions on the real frequency (`w`) axis.
            These functions are stored as piecewise Legendre polynomials.

            To obtain the value of all basis functions at a point or a array of
            points `w`, you can call the function ``v(w)``.  To obtain a single
            basis function, a slice or a subset `l`, you can use ``v[l]``.
    """
    def __init__(self, statistics, beta, wmax, eps=None, *, kernel=None,
                 sve_result=None):
        if not (beta > 0):
            raise ValueError("inverse temperature beta must be positive")
        if not (wmax >= 0):
            raise ValueError("frequency cutoff must be non-negative")

        self.kernel = _get_kernel(statistics, beta * wmax, kernel)
        if sve_result is None:
            u, s, v = sve.compute(self.kernel, eps)
        else:
            u, s, v = sve_result
            if u.shape != s.shape or s.shape != v.shape:
                raise ValueError("mismatched shapes in SVE")

        self.statistics = statistics
        self.beta = beta

        if eps is None and sve_result is None and not sve.HAVE_XPREC:
            warn("xprec package is not available:\n"
                 "expect single precision (1.5e-8) only as both cutoff and\n"
                 "accuracy of the basis functions")

        # The polynomials are scaled to the new variables by transforming the
        # knots according to: tau = beta/2 * (x + 1), w = wmax * y.  Scaling
        # the data is not necessary as the normalization is inferred.
        wmax = self.kernel.lambda_ / self.beta
        self.u = u.__class__(u.data, beta/2 * (u.knots + 1))
        self.v = v.__class__(v.data, wmax * v.knots)

        # The singular values are scaled to match the change of variables, with
        # the additional complexity that the kernel may have an additional
        # power of w.
        self.s = np.sqrt(beta/2 * wmax) * (wmax**(-self.kernel.ypower)) * s

        # HACK: as we don't yet support Fourier transforms on anything but the
        # unit interval, we need to scale the underlying data.  This breaks
        # the correspondence between U.hat and Uhat though.
        uhat_base = u.__class__(np.sqrt(beta) * u.data, u.knots)

        conv_radius = 40 * self.kernel.lambda_
        _even_odd = {'F': 'odd', 'B': 'even'}[statistics]
        self.uhat = uhat_base.hat(_even_odd, conv_radius)

    def __getitem__(self, index):
        """Return basis functions/singular values for given index/indices.

        This can be used to truncate the basis to the n most significant
        singular values: `basis[:3]`.
        """
        sve_result = self.u[index], self.s[index], self.v[index]
        return self.__class__(self.kernel, self.statistics, self.beta,
                              sve_result=sve_result)

    @property
    def wmax(self):
        """Cutoff in real frequency."""
        return self.kernel.lambda_ / self.beta

    @property
    def size(self):
        """Number of basis functions / singular values."""
        return self.u.size

    @property
    def shape(self):
        """Shape of the basis function set"""
        return self.u.shape

    def default_tau_sampling_points(self):
        """Default sampling points on the imaginary time/x axis"""
        return _default_tau_sampling_points(self.u)

    def default_matsubara_sampling_points(self, *, mitigate=True):
        """Default sampling points on the imaginary frequency axis"""
        return _default_matsubara_sampling_points(self.uhat, mitigate)


def _default_tau_sampling_points(u):
    poly = u[-1]
    maxima = poly.deriv().roots()
    left = .5 * (maxima[:1] + poly.xmin)
    right = .5 * (maxima[-1:] + poly.xmax)
    return np.concatenate([left, maxima, right])


def _default_matsubara_sampling_points(uhat, mitigate=True):
    # Use the (discrete) extrema of the corresponding highest-order basis
    # function in Matsubara.  This turns out to be close to optimal with
    # respect to conditioning for this size (within a few percent).
    polyhat = uhat[-1]
    wn = polyhat.extrema()

    # While the condition number for sparse sampling in tau saturates at a
    # modest level, the conditioning in Matsubara steadily deteriorates due
    # to the fact that we are not free to set sampling points continuously.
    # At double precision, tau sampling is better conditioned than iwn
    # by a factor of ~4 (still OK). To battle this, we fence the largest
    # frequency with two carefully chosen oversampling points, which brings
    # the two sampling problems within a factor of 2.
    if mitigate:
        wn_outer = wn[[0, -1]]
        wn_diff = 2 * np.round(0.025 * wn_outer).astype(int)
        if wn.size >= 20:
            wn = np.hstack([wn, wn_outer - wn_diff])
        if wn.size >= 42:
            wn = np.hstack([wn, wn_outer + wn_diff])
        wn = np.unique(wn)

    # For boson, include "0".
    if wn[0] % 2 == 0:
        wn = np.unique(np.hstack((0, wn)))

    return wn


def _get_kernel(statistics, lambda_, kernel):
    if statistics not in 'BF':
        raise ValueError("statistics must either be 'B' (for bosonic basis) "
                         "or 'F' (for fermionic basis)")
    if kernel is None:
        if statistics == 'F':
            kernel = _kernel.KernelFFlat(lambda_)
        else:
            kernel = _kernel.KernelBFlat(lambda_)
    else:
        try:
            lambda_kernel = kernel.lambda_
        except AttributeError:
            pass
        else:
            if not np.allclose(lambda_kernel, lambda_, atol=0, rtol=4e-16):
                raise ValueError("lambda of kernel and basis mismatch")
    return kernel



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
from warnings import warn
import numpy as np
import scipy.linalg.interpolative as intp_decomp

try:
    from xprec import ddouble as _ddouble, finfo
    import xprec.linalg as _xprec_linalg

    MAX_DTYPE = _ddouble
    MAX_EPS = 5e-32
except ImportError:
    _ddouble = None
    _xprec_linalg = None

    MAX_DTYPE = np.double
    MAX_EPS = np.finfo(MAX_DTYPE).eps
    finfo = np.finfo

try:
    from scipy.linalg.lapack import dgejsv as _lapack_dgejsv
except ImportError:
    _lapack_dgejsv = None


def compute(a_matrix, n_sv_hint=None, strategy='fast'):
    """Compute thin/truncated singular value decomposition

    Computes the thin/truncated singular value decomposition of a matrix ``A``
    into ``U``, ``s``, ``V``:

        A == (U * s) @ V.T

    Depending on the strategy, only as few as ``n_sv_hint`` most significant
    singular values may be returned, but applications should not rely on this
    behvaiour.  The ``strategy`` parameter can be ``fast`` (RRQR/t-SVD),
    ``default`` (full SVD) or ``accurate`` (Jacobi rotation SVD).
    """
    a_matrix = np.asarray(a_matrix)
    m, n = a_matrix.shape
    if n_sv_hint is None:
        n_sv_hint = min(m, n)
    n_sv_hint = min(m, n, n_sv_hint)

    if _ddouble is not None and a_matrix.dtype == _ddouble:
        u, s, v = _ddouble_svd_trunc(a_matrix)
    elif strategy == 'fast':
        u, s, v = _idsvd(a_matrix, n_sv=n_sv_hint)
    elif strategy == 'default':
        # Usual (simple) SVD
        u, s, vh = np.linalg.svd(a_matrix, full_matrices=False)
        v = vh.T.conj()
    elif strategy == 'accurate':
        # Most accurate SVD
        if _lapack_dgejsv is None:
            warn("dgejsv (accurate SVD) is not available. Falling back to\n"
                 "default SVD.  Expect slightly lower precision.\n"
                 "Use xprec or scipy >= 1.5 to fix the issue.")
            return compute(a_matrix, n_sv_hint, strategy='default')
        u, s, v = _dgejsv(a_matrix, mode='F')
    else:
        raise ValueError("invalid strategy:" + str(strategy))

    return u, s, v


def _idsvd(a, n_sv):
    # Use interpolative decomposition, since it scales favorably to a full
    # SVD when we are only interested in a small subset of singular values.
    # NOTE: this returns the right singular vectors, not their conjugate!
    intp_decomp.seed(4711)
    return intp_decomp.svd(a, n_sv)


def _dgejsv(a, mode='A'):
    """Compute SVD using the (more accurate) Jacobi method"""
    # GEJSV can only handle tall matrices
    m, n = a.shape
    if m < n:
        u, s, v = _dgejsv(a.T, mode)
        return v, s, u

    mode = mode.upper()
    joba = dict(zip("CEFGAR", range(6)))[mode]
    s, u, v, _stat, istat, info = _lapack_dgejsv(a, joba)
    if info < 0:
        raise ValueError("LAPACK error - invalid parameter")
    if istat[2] != 0:
        warn("a contained denormalized floats - possible loss of accuracy",
             UserWarning, 2)
    if info > 0:
        warn("SVD did not converge", UserWarning, 2)
    return u, s, v


def _ddouble_svd_trunc(a):
    """Truncated SVD with double double precision"""
    if _xprec_linalg is None:
        raise RuntimeError("require xprec package for this precision")
    u, s, vh = _xprec_linalg.svd_trunc(a)
    return u, s, vh.T



"""
Intermediate representation (IR) for many-body propagators
==========================================================
"""
__copyright__ = "Copyright (C) 2020-2021 Markus Wallerberger and others"
__license__ = "MIT"
__version__ = "0.4.4"

min_xprec_version = "1.0"

try:
    import xprec as xprec
    from pkg_resources import parse_version
    from warnings import warn
    if parse_version(xprec.__version__) < parse_version(min_xprec_version):
        warn(f"xprec is too old! Please use xprec>={min_xprec_version}.")
except ImportError:
    pass

from .kernel import KernelFFlat, KernelBFlat
from .sve import compute as compute_sve
from .basis import IRBasis, FiniteTempBasis
from .sampling import TauSampling, MatsubaraSampling



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import sparse_ir

import pytest

BASES = [
    ('F', 73),
    ('B', 37)
    ]


@pytest.fixture(scope="module")
def bases():
    return {(stat, lambda_): sparse_ir.IRBasis(stat, lambda_)
            for (stat, lambda_) in BASES}


def _check_smooth(u, s, uscale, fudge_factor):
    eps = np.finfo(s.dtype).eps
    x = u.knots[1:-1]

    jump = np.abs(u(x + eps) - u(x - eps))
    compare = np.abs(u(x + 3 * eps) - u(x + eps))
    compare = np.maximum(compare, uscale * eps)

    # loss of precision
    compare *= fudge_factor * (s[0] / s)[:, None]
    try:
        np.testing.assert_array_less(jump, compare)
    except:
        print((jump > compare).nonzero())
        raise


@pytest.mark.parametrize("stat,lambda_", BASES)
def test_smooth(bases, stat, lambda_):
    basis = bases[stat, lambda_]
    _check_smooth(basis.u, basis.s, 2*basis.u(1).max(), 24)
    _check_smooth(basis.v, basis.s, 50, 20)


@pytest.mark.parametrize("stat,lambda_", BASES)
def test_num_roots_u(bases, stat, lambda_):
    basis = bases[stat, lambda_]
    for i in range(basis.u.size):
        ui_roots = basis.u[i].roots()
        assert ui_roots.size == i


@pytest.mark.parametrize("stat,lambda_", BASES)
def test_num_roots_uhat(bases, stat, lambda_):
    basis = bases[stat, lambda_]
    for i in [0, 1, 7, 10]:
        x0 = basis.uhat[i].extrema()
        assert i + 1 <= x0.size <= i + 2



import numpy as np
import numpy.polynomial.legendre as np_legendre

from .poly import PiecewiseLegendrePoly
from .basis import _default_matsubara_sampling_points, _default_tau_sampling_points


class LegendreBasis(object):
    r"""Legendre basis

    In the original paper [L. Boehnke et al., PRB 84, 075145 (2011)],
    they used:

        G(\tau) = \sum_{l=0} \sqrt{2l+1} P_l[x(\tau)] G_l/beta,

    where P_l[x] is the $l$-th Legendre polynomial.

    In this class, the basis functions are defined by

        U_l(\tau) \equiv c_l (\sqrt{2l+1}/beta) * P_l[x(\tau)],

    where c_l are additional l-depenent constant factors.
    By default, we take c_l = 1, which reduces to the original definition.
    """
    def __init__(self, statistics, beta, size, cl=None):
        if statistics not in 'BF':
            raise ValueError("Statistics must be either 'B' for bosonic"
                             "or 'F' for fermionic")
        if not (beta > 0):
            raise ValueError("inverse temperature beta must be positive")
        if not (size > 0):
            raise ValueError("size of basis must be positive")

        self.statistics = statistics
        self.beta = beta
        self.size = size
        if cl is None:
            cl = np.ones(self.size)
        self.cl = cl

        # self.u
        knots = np.array([0, beta])
        data = np.zeros((size, knots.size-1, size))
        for l in range(size):
            data[l, 0, l] = np.sqrt((l+0.5)/beta) * cl[l]
        self.u = PiecewiseLegendrePoly(data, knots)

        # self.uhat
        # Hack: See basis.py
        uhat_base = PiecewiseLegendrePoly(np.sqrt(beta) * self.u.data,
                                          np.array([-1,1]))
        self.uhat = uhat_base.hat({'F': 'odd', 'B': 'even'}[statistics])

        # self.v
        self.v = None

    def default_tau_sampling_points(self):
        return 0.5 * self.beta * (np_legendre.leggauss(self.size)[0] + 1)

    def default_matsubara_sampling_points(self, *, mitigate=True):
        return _default_matsubara_sampling_points(self.uhat, mitigate)


class MatsubaraConstBasis(object):
    """Constant term in matsubara-frequency domain

    The unity in the matsubara-frequency domain
    """
    def __init__(self, statistics, beta, value=1):
        if statistics not in 'BF':
            raise ValueError("Statistics must be either 'B' for bosonic"
                            "or 'F' for fermionic")
        if not (beta > 0):
            raise ValueError("inverse temperature beta must be positive")

        self.statistics = statistics
        self.beta = beta
        self.size = 1
        self.value = value

        # self.u
        self.u = None

        # self.uhat
        self.uhat = _ConstTerm(value)

        # self.v
        self.v = None

    def default_tau_sampling_points(self):
        return np.array([])

    def default_matsubara_sampling_points(self, *, mitigate=True):
        return np.array([])


class _ConstTerm:
    def __init__(self, value):
        self.size = 1
        self.value = value

    def __call__(self, n):
        """Return value for given frequencies"""
        return self.value * np.ones_like(n)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import numpy.polynomial.legendre as np_legendre

from sparse_ir import gauss


def test_collocate():
    r = gauss.legendre(20)
    cmat = gauss.legendre_collocation(r)
    emat = np_legendre.legvander(r.x, r.x.size-1)
    np.testing.assert_allclose(emat.dot(cmat), np.eye(20), atol=1e-13, rtol=0)


def _gauss_validate(rule):
    if not (rule.a <= rule.b):
        raise ValueError("a,b must be a valid interval")
    if not (rule.x <= rule.b).all():
        raise ValueError("x must be smaller than b")
    if not (rule.x >= rule.a).all():
        raise ValueError("x must be larger than a")
    if not (rule.x[:-1] < rule.x[1:]).all():
        raise ValueError("x must be well-ordered")
    if rule.x.shape != rule.w.shape:
        raise ValueError("shapes are inconsistent")

    np.testing.assert_allclose(rule.x_forward, rule.x - rule.a)
    np.testing.assert_allclose(rule.x_backward, rule.b - rule.x)

def test_gauss_leg():
    rule = gauss.legendre(200)
    _gauss_validate(rule)
    x, w = np.polynomial.legendre.leggauss(200)
    np.testing.assert_allclose(rule.x, x)
    np.testing.assert_allclose(rule.w, w)


def test_piecewise():
    edges = [-4, -1, 1, 3]
    rule = gauss.legendre(20).piecewise(edges)
    _gauss_validate(rule)



import numpy as np


class _CompositeBasisFunctionBase:
    """Union of several basis functions"""
    def __init__(self, polys):
        self._polys = tuple(polys)
        self._sizes = np.array([p.size for p in self._polys])
        self._cumsum_sizes = np.cumsum(self._sizes)
        self._size = np.sum(self._sizes)

    def __getitem__(self, l):
        """Return part of a set of basis functions"""
        if isinstance(l, int) or issubclass(type(l), np.integer):
            idx_p, l_ = self._internal_pos(l)
            return self._polys[idx_p][l_]
        else:
            raise ValueError(f"Unsupported type of l {type(l)}!")

    def _internal_pos(self, l):
        idx_p = np.searchsorted(self._cumsum_sizes, l, "right")
        l_ = l - self._cumsum_sizes[idx_p-1] if idx_p >= 1 else l
        return idx_p, l_

    @property
    def shape(self): return (self.size,)

    @property
    def size(self): return self._size


class CompositeBasisFunction(_CompositeBasisFunctionBase):
    """Union of several basis functions for the imaginary-time/real-frequency
    domains"""
    def __init__(self, polys):
        """Initialize CompositeBasisFunction

        Arguments:
        ----------
         - polys: iterable object of basis-function-like instances
        """
        super().__init__(polys)

    def __call__(self, x):
        """Evaluate basis function at position x"""
        return np.vstack([p(x) for p in self._polys])

    def value(self, l, x):
        """Return value for l and x."""
        if not isinstance(l, np.ndarray):
            l = np.asarray(l)
        if not isinstance(x, np.ndarray):
            x = np.asarray(x)
        return np.squeeze(np.asarray([self[ll](xx) for ll, xx in zip(l, x)]))

    def overlap(self, f, axis=None, _deg=None):
        r"""Evaluate overlap integral of this basis function with function `f`"""
        return np.vstack((p.overlap(f, axis, _deg) for p in self._polys))

    def deriv(self, n=1):
        """Get the n'th derivative"""
        return np.vstack([p.deriv(n) for p in self._polys])

    #def hat(self, freq, n_asymp=None):
        #"""Get Fourier transformed object"""
        #return CompositeBasisFunctionFT([p.hat(freq, n_asymp) for p in self._polys])

    #def roots(self, alpha=2):
        #"""Find all roots of the basis function """
        #return np.unique(np.hstack([p.roots(alpha) for p in self._polys]))


class CompositeBasisFunctionFT(_CompositeBasisFunctionBase):
    """Union of fourier transform of several basis functions for the
    Matsubara domain"""
    def __init__(self, polys):
        """Initialize CompositeBasisFunctionFT

        Arguments:
        ----------
         - polys: iterable object of basis-function-like instances
        """
        super().__init__(polys)

    def __call__(self, n):
        """Obtain Fourier transform of basis function for given frequencies"""
        return np.vstack([p(n) for p in self._polys])


class CompositeBasis:
    """Union of several basis sets"""
    def __init__(self, bases):
        """Initialize a composite basis

        Args:
            bases): iterable object of FiniteTempBasis instances
        """
        if not all(b.statistics == bases[0].statistics for b in bases):
            raise ValueError("All bases must have the same statistics!")
        if not all(b.beta == bases[0].beta for b in bases):
            raise ValueError("All bases must have the same beta!")

        self._size = np.sum([b.size for b in bases])
        self.bases = bases
        self.u = CompositeBasisFunction([b.u for b in bases]) \
                    if all(b.u is not None for b in bases) else None
        self.v = CompositeBasisFunction([b.v for b in bases]) \
                    if all(b.v is not None for b in bases) else None
        self.uhat = CompositeBasisFunctionFT([b.uhat for b in bases]) \
                    if all(b.uhat is not None for b in bases) else None

    @property
    def size(self): return self._size

    @property
    def shape(self): return (self._size,)

    def default_tau_sampling_points(self):
        return np.unique(np.hstack(
                    [b.default_tau_sampling_points() for b in self.bases]))

    def default_matsubara_sampling_points(self, *, mitigate=True):
        return np.unique(np.hstack(
                    [b.default_matsubara_sampling_points(mitigate=mitigate)
                     for b in self.bases]))



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import pytest
import importlib

from sparse_ir.basis import FiniteTempBasis
from sparse_ir.kernel import KernelFFlat, KernelBFlat

eps = None
if importlib.util.find_spec("xprec") is not None:
    eps = 1.0e-15

all_basis_sets = [(stat, lambda_)
                  for stat in 'FB' for lambda_ in [1E+1, 1E+3]]

"""
A pole at omega=pole. Compare analytic results of G(iwn) and numerical
results computed by using unl.
"""
@pytest.mark.parametrize("stat, lambda_", all_basis_sets)
def test_single_pole(stat, lambda_):
    wmax = 1.0
    pole = 0.1 * wmax
    beta = lambda_/wmax
    basis = FiniteTempBasis(stat, beta, wmax, eps)

    if stat == 'F':
        stat_shift = 1
    else:
        stat_shift = 0
    rho_l = basis.v(pole/wmax) * pole**(-basis.kernel.ypower)
    gl = - basis.s * rho_l

    func_G = lambda n: 1/(1J * (2*n+stat_shift)*np.pi/beta - pole)

    # Compute G(iwn) using unl
    matsu_test = np.array([-1, 0, 1, 1E+2, 1E+4, 1E+6, 1E+8, 1E+10, 1E+12],
                          dtype=np.int64)
    prj_w = basis.uhat(2*matsu_test+stat_shift).T
    Giwn_t = prj_w @ gl

    # Compute G(iwn) from analytic expression
    Giwn_ref = func_G(matsu_test)

    magnitude = np.abs(Giwn_ref).max()
    diff = np.abs(Giwn_t - Giwn_ref)

    tol = max(10 * basis.s[-1]/basis.s[0], 1e-10)

    # Absolute error
    assert (diff/magnitude).max() < tol

    # Relative error
    #print("relative: error ", np.abs(diff/Giwn_ref))
    assert np.amax(np.abs(diff/Giwn_ref)) < tol



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import sparse_ir
from sparse_ir import augment
from scipy.special import eval_legendre, spherical_jn

import pytest

def compute_Tnl(vsample, n_legendre):
    """
    Compute transformation matrix from Legendre to fermionic/bosonic Matsubara frequency
    Implement Eq. (4.5) in the Boehnke's  Ph.D thesis
    """
    Tnl = np.zeros((vsample.size, n_legendre), dtype=np.complex128)
    for idx_n, v in enumerate(vsample):
        abs_v = abs(v)
        sph_jn = np.array(
            [spherical_jn(l, 0.5*abs_v*np.pi) for l in range(n_legendre)])
        for il in range(n_legendre):
            Tnl[idx_n, il] = (1J**(abs_v+il)) * np.sqrt(2*il + 1.0) * sph_jn[il]
        if v < 0:
            Tnl[idx_n, :] = Tnl[idx_n, :].conj()
    return Tnl

@pytest.mark.parametrize("stat", ["F", "B"])
def test_legendre_basis(stat):
    beta = 10.0
    Nl = 100
    cl = np.sqrt(2*np.arange(Nl) + 1)
    basis = augment.LegendreBasis(stat, beta, Nl, cl=cl)

    tau = np.array([0, 0.1*beta, 0.4*beta, beta])
    uval = basis.u(tau)
    l = 0
    for l in range(Nl):
        np.testing.assert_allclose(
            uval[l,:],
            cl[l] * (np.sqrt(2*l+1)/beta) * eval_legendre(l, 2*tau/beta-1)
            )

    zeta = {"F":  1, "B": 0}[stat]
    sign = {"F": -1, "B": 1}[stat]
    w = 2*np.arange(-10, 10) + zeta

    uhat_val = basis.uhat(w)
    uhat_val_ref = compute_Tnl(w, Nl)
    np.testing.assert_allclose(uhat_val.T, uhat_val_ref * cl[None,:])

    # G(iv) = 1/(iv-pole)
    # G(tau) = -e^{-tau*pole}/(1 + e^{-beta*pole}) [F]
    #        = -e^{-tau*pole}/(1 - e^{-beta*pole}) [B]
    pole = 1.0
    tau_smpl = sparse_ir.TauSampling(basis)
    gtau = -np.exp(-tau_smpl.sampling_points * pole)/(1 - sign * np.exp(-beta * pole))
    gl_from_tau = tau_smpl.fit(gtau)

    matsu_smpl = sparse_ir.MatsubaraSampling(basis)
    giv = 1/(1J*matsu_smpl.sampling_points*np.pi/beta - pole)
    gl_from_matsu = matsu_smpl.fit(giv)

    np.testing.assert_allclose(gl_from_tau, gl_from_matsu,
                               atol=1e-10*np.abs(gl_from_matsu).max(), rtol=0)


@pytest.mark.parametrize("stat", ["F", "B"])
def test_const_basis(stat):
    beta = 100
    value = 2.001
    basis = augment.MatsubaraConstBasis(stat, beta, value)

    zeta = {"F":  1, "B": 0}[stat]
    v = 2*np.arange(10)+zeta
    np.testing.assert_array_equal(basis.uhat(v), np.ones_like(v) * value)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import io, os.path, re
from setuptools import setup, find_packages


def readfile(*parts):
    """Return contents of file with path relative to script directory"""
    herepath = os.path.abspath(os.path.dirname(__file__))
    fullpath = os.path.join(herepath, *parts)
    with io.open(fullpath, 'r') as f:
        return f.read()


def extract_varval(varname):
    """Extract value of __version__ variable by parsing python script"""
    initfile = readfile('src', 'sparse_ir', '__init__.py')
    var_re = re.compile(rf"(?m)^{varname}\s*=\s*['\"]([^'\"]*)['\"]")
    match = var_re.search(initfile)
    return match.group(1)


REPO_URL = "https://github.com/SpM-lab/sparse-ir"
VERSION = extract_varval('__version__')
MIN_XPREC_VERSION = extract_varval('min_xprec_version')
LONG_DESCRIPTION = readfile('README.rst')

setup(
    name='sparse-ir',
    version=VERSION,

    description=
        'intermediate representation (IR) basis for electronic propagator',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    keywords=' '.join([
        'irbasis'
        ]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        ],

    url=REPO_URL,
    author=[
        'Markus Wallerberger',
        'Hiroshi Shinaoka',
        'Kazuyoshi Yoshimi',
        'Junya Otsuki',
        'Chikano Naoya'
        ],
    author_email='markus.wallerberger@tuwien.ac.at',

    python_requires='>=3',
    install_requires=[
        'numpy',
        'scipy',
        'setuptools'
    ],
    extras_require={
        'test': ['pytest', 'irbasis', 'xprec'],
        'doc': ['sphinx>=2.1', 'sphinx_rtd_theme'],
        'xprec': [f'xprec>={MIN_XPREC_VERSION}'],
        },

    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    )



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
"""
Auxiliary module for root finding routines.
"""
import numpy as np


def find_all(f, xgrid, type='continuous'):
    """Find all roots of function between gridpoints"""
    xgrid = np.asarray(xgrid)
    if xgrid.ndim != 1:
        raise ValueError("grid must be a one-dimensional array")

    # First, extract roots that lie directly on the grid points
    fx = f(xgrid)
    hit = fx == 0
    x_hit = xgrid[hit]

    # Next, find out where the sign changes (sign bit flips) remove the
    # previously found points from consideration (we need to remove both
    # directions for transitions + -> - and - -> +)
    sign_change = np.signbit(fx[:-1]) != np.signbit(fx[1:])
    sign_change &= ~hit[:-1] & ~hit[1:]
    if not sign_change.any():
        return x_hit

    # sign_change[i] being set means that the sign changes from xgrid[i] to
    # xgrid[i+1].  This means a corresponds to those xgrid[i] and b to those
    # xgrid[i+1] where sign_change[i] is set.
    where_a = np.hstack((sign_change, False))
    where_b = np.hstack((False, sign_change))
    a = xgrid[where_a]
    b = xgrid[where_b]
    fa = fx[where_a]
    fb = fx[where_b]

    # Depending on whether we have a discrete or continuous function, do
    # this.
    if type == 'continuous':
        xeps = np.finfo(xgrid.dtype).eps * np.abs(xgrid).max()
        x_bisect = _bisect_cont(f, a, b, fa, fb, xeps)
    elif type == 'discrete':
        x_bisect = _bisect_discr(f, a, b, fa, fb)
    else:
        raise ValueError("invalid type")
    return np.sort(np.hstack([x_hit, x_bisect]))


def _bisect_cont(f, a, b, fa, fb, xeps):
    """Bisect roots already found"""
    while True:
        mid = 0.5 * (a + b)
        fmid = f(mid)
        towards_a = np.signbit(fa) != np.signbit(fmid)
        a = np.where(towards_a, a, mid)
        fa = np.where(towards_a, fa, fmid)
        b = np.where(towards_a, mid, b)
        fb = np.where(towards_a, fmid, fb)
        found = b - a < xeps
        if found.any():
            break

    roots = mid[found]
    if found.all():
        return roots
    more = _bisect_cont(f, a[~found], b[~found], fa[~found], fb[~found], xeps)
    return np.hstack([roots, more])


def _bisect_discr(f, a, b, fa, fb):
    """Bisect roots already found"""
    while True:
        mid = (a + b) // 2
        found = a == mid
        if found.any():
            break

        fmid = f(mid)
        towards_a = np.signbit(fa) != np.signbit(fmid)
        a = np.where(towards_a, a, mid)
        fa = np.where(towards_a, fa, fmid)
        b = np.where(towards_a, mid, b)
        fb = np.where(towards_a, fmid, fb)

    roots = mid[found]
    if found.all():
        return roots
    more = _bisect_discr(f, a[~found], b[~found], fa[~found], fb[~found])
    return np.hstack([roots, more])


def discrete_extrema(f, xgrid):
    """Find extrema of Bessel-like discrete function"""
    fx = f(xgrid)
    absfx = np.abs(fx)

    # Forward differences: where[i] now means that the secant changes sign
    # fx[i+1]. This means that the extremum is STRICTLY between x[i] and
    # x[i+2]
    gx = fx[1:] - fx[:-1]
    sgx = np.signbit(gx)
    where = sgx[:-1] != sgx[1:]
    where_a = np.hstack([where, False, False])
    where_b = np.hstack([False, False, where])

    a = xgrid[where_a]
    b = xgrid[where_b]
    absf_a = absfx[where_a]
    absf_b = absfx[where_b]
    res = [_bisect_discr_extremum(f, *args)
           for args in zip(a, b, absf_a, absf_b)]

    # We consider the outer point to be extremua if there is a decrease
    # in magnitude or a sign change inwards
    sfx = np.signbit(fx)
    if absfx[0] > absfx[1] or sfx[0] != sfx[1]:
        res.insert(0, xgrid[0])
    if absfx[-1] > absfx[-2] or sfx[-1] != sfx[-2]:
        res.append(xgrid[-1])

    return np.array(res)


def _bisect_discr_extremum(f, a, b, absf_a, absf_b):
    """Bisect extremum of f on the set {a+1, ..., b-1}"""
    d = b - a
    if d <= 1:
        return a if absf_a > absf_b else b
    if d == 2:
        return a + 1

    m = (a + b) // 2
    n = m + 1
    absf_m = np.abs(f(m))
    absf_n = np.abs(f(n))
    if absf_m > absf_n:
        return _bisect_discr_extremum(f, a, n, absf_a, absf_n)
    else:
        return _bisect_discr_extremum(f, m, b, absf_m, absf_b)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import pytest

from sparse_ir import adapter

try:
    import irbasis
except ImportError:
    pytest.skip("no irbasis library for comparison", allow_module_level=True)
    raise

COMPARE_PARAMS = [
    ('F', 10),
    ('F', 10_000),
    ('B', 10),
    ('B', 10_000),
    ]

# Return of fixture functions are passed as parameters to test functions
# if the parameter has the same name as the fixture function.  scope="module"
# ensures that fixtures are cached.

@pytest.fixture(scope="module")
def adapters():
    return {params: adapter.load(*params) for params in COMPARE_PARAMS}


@pytest.fixture(scope="module")
def references():
    return {params: irbasis.load(*params) for params in COMPARE_PARAMS}


@pytest.mark.parametrize("stat,lambda_", COMPARE_PARAMS)
def test_svals(stat, lambda_, adapters, references):
    adapt = adapters[stat, lambda_]
    ref = references[stat, lambda_]
    shared = slice(min(adapt.dim(), ref.dim()))
    eps = np.finfo(float).eps

    assert adapt.statistics == ref.statistics == stat
    assert adapt.Lambda == ref.Lambda == lambda_
    assert adapt.dim() > 10

    np.testing.assert_allclose(adapt.sl(shared), ref.sl(shared),
                               atol=10 * ref.sl(0) * eps, rtol=0)

@pytest.mark.parametrize("stat,lambda_", COMPARE_PARAMS)
def test_vly(stat, lambda_, adapters, references):
    adapt = adapters[stat, lambda_]
    ref = references[stat, lambda_]
    shared_dim = min(adapt.dim(), ref.dim())

    y = [-1., -.5, -.1, -.01, -.001, -.0001, 0, .0001, .001, .01, .1, .5, 1.]
    for li in [0, 1, 2, shared_dim//2, shared_dim//2 + 1, shared_dim-1]:
        uly_adapt = adapt.vly(li, y)
        uly_ref = ref.vly(li, y)
        tol = 1e-10 * np.abs(uly_ref).max() * ref.sl(0) / ref.sl(li)
        np.testing.assert_allclose(uly_adapt, uly_ref, atol=tol, rtol=0)


@pytest.mark.parametrize("stat,lambda_", COMPARE_PARAMS)
def test_ulx(stat, lambda_, adapters, references):
    adapt = adapters[stat, lambda_]
    ref = references[stat, lambda_]
    shared_dim = min(adapt.dim(), ref.dim())

    x = [-1., -.9999, -.999, -.99, -.9, -.5, 0., .5, .9, .99, .999, .9999, 1.]
    for li in [0, 1, 2, shared_dim//2, shared_dim//2 + 1, shared_dim-1]:
        uly_adapt = adapt.ulx(li, x)
        uly_ref = ref.ulx(li, x)
        tol = 1e-12 * np.abs(uly_ref).max() * ref.sl(0) / ref.sl(li)
        np.testing.assert_allclose(uly_adapt, uly_ref, atol=tol, rtol=0)


@pytest.mark.parametrize("stat,lambda_", COMPARE_PARAMS)
def test_ulx(stat, lambda_, adapters, references):
    adapt = adapters[stat, lambda_]
    ref = references[stat, lambda_]
    shared_dim = min(adapt.dim(), ref.dim())

    n = [-20, -2, -1, 0, 1, 2, 20, 100, 300, 1000]
    for li in [0, 1, 2, shared_dim//2, shared_dim//2 + 1, shared_dim-1]:
        ulxhat_adapt = adapt.compute_unl(n, li)
        ulxhat_ref = ref.compute_unl(n, li).ravel()
        tol = 1e-13 * np.abs(ulxhat_ref).max() * ref.sl(0) / ref.sl(li)
        np.testing.assert_allclose(ulxhat_adapt, ulxhat_ref, atol=tol, rtol=0)



# Copyright (C) 2020-2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
from warnings import warn
import numpy as np

from . import gauss
from . import poly
from . import svd
from . import kernel

HAVE_XPREC = svd._ddouble is not None


def compute(K, eps=None, n_sv=None, n_gauss=None, dtype=float, work_dtype=None,
            sve_strat=None, svd_strat=None):
    """Perform truncated singular value expansion of a kernel.

    Perform a truncated singular value expansion (SVE) of a integral
    kernel ``K : [xmin, xmax] x [ymin, ymax] -> R``:

        K(x, y) == sum(s[l] * u[l](x) * v[l](y) for l in (0, 1, 2, ...)),

    where ``s[l]`` are the singular values, which are ordered in non-increasing
    fashion, ``u[l](x)`` are the left singular functions, which form an
    orthonormal system on ``[xmin, xmax]``, and ``v[l](y)`` are the right
    singular functions, which for an orthonormal system on ``[ymin, ymax]``

    The SVE is mapped onto the singular value decomposition (SVD) of a matrix
    by expanding the kernel in piecewise Legendre polynomials (by default by
    using a collocation).

    Arguments:

      - ``K``: Integral kernel to take SVE from
      - ``eps``:  Relative cutoff for the singular values.  Defaults to double
        precision (2.2e-16) if the xprec package is available, and 1.5e-8
        otherwise.
      - ``n_sv``: Maximum basis size.  If given, only at most the ``n_sv`` most
        significant singular values and associated singular functions are
        returned.
      - ``n_gauss``: Order of Legendre polynomials.  Defaults to hinted value
        by the kernel.
      - ``dtype``: Data type of the result.
      - ``work_dtype``: Working data type.  Defaults to a data type with
        machine epsilon of at least ``eps**2``, or otherwise most accurate data
        type available.
      - ``sve_strat``: SVE to SVD translation strategy.  Defaults to SamplingSVE.
      - ``svd_strat``: SVD solver. Defaults to fast (ID/RRQR) based solution
         when accuracy goals are moderate, and more accurate Jacobi-based
         algorithm otherwise.

    Return tuple ``(u, s, v)``, where:

     - ``u`` is a ``PiecewisePoly`` instance holding the left singular functions
     - ``s`` is a vector of singular values
     - ``v`` is a ``PiecewisePoly`` instance holding the right singular functions
    """
    if eps is None or work_dtype is None or svd_strat is None:
        eps, work_dtype, default_svd_strat = _choose_accuracy(eps, work_dtype)
    if svd_strat is None:
        svd_strat = default_svd_strat
    if sve_strat is None:
        sve_strat = CentrosymmSVE if K.is_centrosymmetric else SamplingSVE
    sve = sve_strat(K, eps, n_gauss=n_gauss, dtype=work_dtype)
    u, s, v = zip(*(svd.compute(matrix, sve.nsvals_hint, svd_strat)
                    for matrix in sve.matrices))
    u, s, v = truncate(u, s, v, eps, n_sv)
    return sve.postprocess(u, s, v, dtype)


class SamplingSVE:
    """SVE to SVD translation by sampling technique [1].

    Maps the singular value expansion (SVE) of a kernel ``K`` onto the singular
    value decomposition of a matrix ``A``.  This is achieved by chosing two
    sets of Gauss quadrature rules: ``(x, wx)`` and ``(y, wy)`` and
    approximating the integrals in the SVE equations by finite sums.  This
    implies that the singular values of the SVE are well-approximated by the
    singular values of the following matrix:

        A[i, j] = sqrt(wx[i]) * K(x[i], y[j]) * sqrt(wy[j])

    and the values of the singular functions at the Gauss sampling points can
    be reconstructed from the singular vectors ``u`` and ``v`` as follows:

        u[l,i] ≈ sqrt(wx[i]) u[l](x[i])
        v[l,j] ≈ sqrt(wy[j]) u[l](y[j])

    [1] P. Hansen, Discrete Inverse Problems, Ch. 3.1
    """
    def __init__(self, K, eps, *, n_gauss=None, dtype=float):
        self.K = K
        sve_hints = K.sve_hints(eps)
        if n_gauss is None:
            n_gauss = sve_hints.ngauss

        self.eps = eps
        self.n_gauss = n_gauss
        self.nsvals_hint = sve_hints.nsvals
        self._rule = gauss.legendre(n_gauss, dtype)
        self._segs_x = sve_hints.segments_x.astype(dtype)
        self._segs_y = sve_hints.segments_y.astype(dtype)
        self._gauss_x = self._rule.piecewise(self._segs_x)
        self._gauss_y = self._rule.piecewise(self._segs_y)
        self._sqrtw_x = np.sqrt(self._gauss_x.w)
        self._sqrtw_y = np.sqrt(self._gauss_y.w)

    @property
    def matrices(self):
        """SVD problems underlying the SVE."""
        result = kernel.matrix_from_gauss(self.K, self._gauss_x, self._gauss_y)
        result *= self._sqrtw_x[:, None]
        result *= self._sqrtw_y[None, :]
        return result,

    def postprocess(self, u, s, v, dtype=None):
        """Constructs the SVE result from the SVD"""
        if dtype is None:
            dtype = np.result_type(u, s, v)

        s = s.astype(dtype)
        u_x = u / self._sqrtw_x[:,None]
        v_y = v / self._sqrtw_y[:,None]

        u_x = u_x.reshape(self._segs_x.size - 1, self.n_gauss, s.size)
        v_y = v_y.reshape(self._segs_y.size - 1, self.n_gauss, s.size)

        cmat = gauss.legendre_collocation(self._rule)
        # lx,ixs -> ils -> lis
        u_data = (cmat @ u_x).transpose(1, 0, 2)
        v_data = (cmat @ v_y).transpose(1, 0, 2)

        dsegs_x = self._segs_x[1:] - self._segs_x[:-1]
        dsegs_y = self._segs_y[1:] - self._segs_y[:-1]
        u_data *= np.sqrt(.5 * dsegs_x)[None,:,None]
        v_data *= np.sqrt(.5 * dsegs_y)[None,:,None]

        # Construct polynomial
        ulx = poly.PiecewiseLegendrePoly(
                        u_data.astype(dtype), self._segs_x.astype(dtype))
        vly = poly.PiecewiseLegendrePoly(
                        v_data.astype(dtype), self._segs_y.astype(dtype))
        _canonicalize(ulx, vly)
        return ulx, s, vly


class CentrosymmSVE:
    """SVE of centrosymmetric kernel in block-diagonal (even/odd) basis.

    For a centrosymmetric kernel ``K``, i.e., a kernel satisfying:
    ``K(x, y) == K(-x, -y)``, one can make the following ansatz for the
    singular functions:

        u[l](x) = ured[l](x) + sign[l] * ured[l](-x)
        v[l](y) = vred[l](y) + sign[l] * ured[l](-y)

    where ``sign[l]`` is either +1 or -1.  This means that the singular value
    expansion can be block-diagonalized into an even and an odd part by
    (anti-)symmetrizing the kernel:

        Keven = K(x, y) + K(x, -y)
        Kodd  = K(x, y) - K(x, -y)

    The l-th basis function, restricted to the positive interval, is then
    the singular function of one of these kernels.  If the kernel generates a
    Chebyshev system [1], then even and odd basis functions alternate.

    [1]: A. Karlin, Total Positivity (1968).
    """
    def __init__(self, K, eps, *, InnerSVE=None, **inner_args):
        if InnerSVE is None:
            InnerSVE = SamplingSVE
        self.K = K
        self.eps = eps

        # Inner kernels for even and odd functions
        self.even = InnerSVE(K.get_symmetrized(+1), eps, **inner_args)
        self.odd = InnerSVE(K.get_symmetrized(-1), eps, **inner_args)

        # Now extract the hints
        self.nsvals_hint = max(self.even.nsvals_hint, self.odd.nsvals_hint)

    @property
    def matrices(self):
        """Set of SVD problems underlying the SVE."""
        m, = self.even.matrices
        yield m
        m, = self.odd.matrices
        yield m

    def postprocess(self, u, s, v, dtype):
        """Constructs the SVE result from the SVD"""
        u_even, s_even, v_even = self.even.postprocess(u[0], s[0], v[0], dtype)
        u_odd, s_odd, v_odd = self.odd.postprocess(u[1], s[1], v[1], dtype)

        # Merge two sets - data is [legendre, segment, l]
        u_data = np.concatenate([u_even.data, u_odd.data], axis=2)
        v_data = np.concatenate([v_even.data, v_odd.data], axis=2)
        s = np.concatenate([s_even, s_odd])
        signs = np.concatenate([np.ones(s_even.size), -np.ones(s_odd.size)])

        # Sort: now for totally positive kernels like defined in this module,
        # this strictly speaking is not necessary as we know that the even/odd
        # functions intersperse.
        sort = s.argsort()[::-1]
        u_data = u_data[:, :, sort]
        v_data = v_data[:, :, sort]
        s = s[sort]
        signs = signs[sort]

        # Extend to the negative side
        inv_sqrt2 = 1/np.sqrt(2, dtype=u_data.dtype)
        u_data *= inv_sqrt2
        v_data *= inv_sqrt2
        poly_flip_x = ((-1)**np.arange(u_data.shape[0]))[:, None, None]
        u_neg = u_data[:, ::-1, :] * poly_flip_x * signs
        v_neg = v_data[:, ::-1, :] * poly_flip_x * signs
        u_data = np.concatenate([u_neg, u_data], axis=1)
        v_data = np.concatenate([v_neg, v_data], axis=1)

        # TODO: this relies on specific symmetrization behaviour ...
        full_hints = self.K.sve_hints(self.eps)
        u = poly.PiecewiseLegendrePoly(u_data, full_hints.segments_x)
        v = poly.PiecewiseLegendrePoly(v_data, full_hints.segments_y)
        return u, s, v


def _choose_accuracy(eps, work_dtype):
    """Choose work dtype and accuracy based on specs and defaults"""
    if eps is None:
        if work_dtype is None:
            return np.sqrt(svd.MAX_EPS), svd.MAX_DTYPE, 'fast'
        safe_eps = np.sqrt(svd.finfo(work_dtype).eps)
        return safe_eps, work_dtype, 'fast'

    if work_dtype is None:
        if eps >= np.sqrt(svd.finfo(float).eps):
            return eps, float, 'fast'
        work_dtype = svd.MAX_DTYPE

    safe_eps = np.sqrt(svd.finfo(work_dtype).eps)
    if eps >= safe_eps:
        svd_strat = 'fast'
    else:
        svd_strat = 'accurate'
        msg = ("\nBasis cutoff is {:.2g}, which is below sqrt(eps) with\n"
               "eps = {:.2g}.  Expect singular values and basis functions\n"
               "for large l to have lower precision than the cutoff.\n")
        msg = msg.format(float(eps), float(np.square(safe_eps)))
        if not HAVE_XPREC:
            msg += "You can install the xprec package to gain more precision.\n"
        warn(msg, UserWarning, 3)

    return eps, work_dtype, svd_strat


def _canonicalize(ulx, vly):
    """Canonicalize basis.

    Each SVD (u_l, v_l) pair is unique only up to a global phase, which may
    differ from implementation to implementation and also platform.  We
    fix that gauge by demanding u_l(1) > 0.  This ensures a diffeomorphic
    connection to the Legendre polynomials for lambda_ -> 0.
    """
    gauge = np.sign(ulx(1))
    ulx.data[None, None, :] *= 1/gauge
    vly.data[None, None, :] *= gauge


def truncate(u, s, v, rtol=0, lmax=None):
    """Truncate singular value expansion.

    Arguments:

     - ``u``, ``s``, ``v``: Thin singular value expansion
     - ``rtol`` : If given, only singular values satisfying
       ``s[l]/s[0] > rtol`` are retained.
     - ``lmax`` : If given, at most the ``lmax`` most significant singular
       values are retained.
    """
    if lmax is not None and (lmax < 0 or int(lmax) != lmax):
        raise ValueError("invalid value of maximum number of singular values")
    if rtol < 0 or rtol > 1:
        raise ValueError("invalid relative tolerance")

    sall = np.hstack(s)

    # Determine singular value cutoff.  Note that by selecting a cutoff even
    # in the case of lmax, we make sure to never remove parts of a degenerate
    # singular value space, rather, we reduce the size of the basis.
    ssort = np.sort(sall)
    cutoff = rtol * ssort[-1]
    if lmax is not None and lmax < sall.size:
        cutoff = max(cutoff, ssort[sall.size - lmax - 1])

    # Determine how many singular values survive in each group
    swhich = np.repeat(np.arange(len(s)), [si.size for si in s])
    scount = np.bincount(swhich, sall > cutoff, minlength=len(s)).astype(int)

    u_cut = [ui[:, :counti] for (ui, counti) in zip(u, scount)]
    s_cut = [si[:counti] for (si, counti) in zip(s, scount)]
    v_cut = [vi[:, :counti] for (vi, counti) in zip(v, scount)]
    return u_cut, s_cut, v_cut
